from django.contrib import admin

# Register your models here.
from .models.following import Following, FollowRequest
from .models.profile import Profile
from .models.project import Project
from .models.event import Event
from .models.file import File
from .models.notification import Notification
from .models.collaboration_invite import CollaborationInvite
from .models.collaboration_request import CollaborationRequest
from .models.comment import Comment
from .models.rating import Rating
from .models.tag import Tag
from .models.report import Report

admin.site.register(Profile)
admin.site.register(Project)
admin.site.register(Event)
admin.site.register(File)
admin.site.register(Following)
admin.site.register(FollowRequest)
admin.site.register(Notification)
admin.site.register(CollaborationInvite)
admin.site.register(CollaborationRequest)
admin.site.register(Comment)
admin.site.register(Rating)
admin.site.register(Tag)
admin.site.register(Report)



# Generated by Django 3.1.4 on 2020-12-21 20:47

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0015_auto_20201219_1124'),
    ]

    operations = [
        migrations.CreateModel(
            name='Rating',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('rating', models.IntegerField(validators=[django.core.validators.MaxValueValidator(10), django.core.validators.MinValueValidator(0)])),
                ('from_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ratings_made', to=settings.AUTH_USER_MODEL)),
                ('to_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ratings_received', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('comment', models.CharField(max_length=200)),
                ('from_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments_made', to=settings.AUTH_USER_MODEL)),
                ('to_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments_received', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.AddConstraint(
            model_name='rating',
            constraint=models.UniqueConstraint(fields=('from_user', 'to_user'), name='unique_rating'),
        ),
    ]



# Generated by Django 3.1.4 on 2021-01-07 21:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0017_auto_20201228_1217'),
        ('api', '0016_auto_20201221_2147'),
    ]

    operations = [
    ]



from rest_framework import serializers
from api.models.report import Report


class ReportSerializer(serializers.ModelSerializer):
    """
    Report serializer
    """

    class Meta:
        model = Report
        fields = '__all__'



# Generated by Django 3.1.3 on 2020-12-08 17:42

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0008_auto_20201206_2028'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='state',
            field=models.CharField(choices=[('draft', '0'), ('inviting collaborators', '1'), ('open for collaborators', '2'), ('in progress', '3'), ('submitted to event', '4'), ('published', '5'), ('cancelled', '6'), ('done', '7'), ('reopened', '8')], default='draft', max_length=100),
        ),
        migrations.CreateModel(
            name='CollaborationInvite',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.CharField(blank=True, default='', max_length=200)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('rejected', models.DateTimeField(blank=True, null=True)),
                ('from_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='collab_invite_sent', to=settings.AUTH_USER_MODEL)),
                ('to_project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='collab_invite_of_project', to='api.project')),
                ('to_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='collab_invite_received', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
    ]



from rest_framework import permissions
from drf_yasg.views import get_schema_view
from drf_yasg import openapi

"""
Swagger configuration
"""
schema_view = get_schema_view(
    openapi.Info(
        title="PaperLayer API",
        default_version='v1',
        description="API documentation of PaperLayer",

    ),
    public=True,
    permission_classes=(permissions.AllowAny,),
)



from rest_framework import permissions


class IsOwnerOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow owners of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        # Write permissions are only allowed to the owner of the snippet.
        return obj.owner == request.user


class IsMemberOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow members of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        # Write permissions are only allowed to the members of the snippet.
        return request.user in obj.members.all() \
            or request.user == obj.owner


class IsRequestSenderOrReceiver(permissions.BasePermission):
    """
    Custom permission to only allow users to send request.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return request.user.id == obj.req_from_user.id \
                or request.user.id == obj.req_to_user.id

        # Write permissions are only allowed to the members of the snippet.
        return request.user.id == obj.req_from_user


class IsMilestoneMemberOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow members of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        # Write permissions are only allowed to the members of the project.
        return request.user in obj.project.members.all() or \
            request.user == obj.project.owner


class IsFileMember(permissions.BasePermission):
    """
    Custom permission to only allow members of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        safe_method = request.method in permissions.SAFE_METHODS
        public = obj.project.is_public

        if safe_method and public:
            return True

        # Write permissions are only allowed to the members of the project.
        return request.user in obj.project.members.all() or \
            request.user == obj.project.owner


class CollaborationPermissions(permissions.BasePermission):
    """
    Custom permission to only allow members of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        if view.action in ['retrieve', 'list']:
            return request.user == obj.to_user or \
                request.user == obj.from_user
        elif view.action in ['update', 'partial_update', 'destroy']:
            return request.user == obj.from_user
        return True


class ProfileDeletion(permissions.BasePermission):
    """
    Custom permission to only allow members of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        if view.action in ['destroy']:
            if request.user.is_staff:
                return True
            else:
                return False

        return True


class CommentPermission(permissions.BasePermission):
    """
    Custom permission to only allow owner of a comment edit it.
    Commented user can also delete the comment.
    """

    def has_object_permission(self, request, view, obj):
        if view.action in ['destroy', 'update', 'partial_update']:
            return request.user == obj.from_user

        return True


class RatingPermission(permissions.BasePermission):
    """
    Custom permission to only allow owner of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        return request.user == obj.from_user



from django.contrib.auth.models import User
from rest_framework import serializers

from api.serializers.profile import ProfileFullSerializer, \
    ProfileBasicSerializer, ProfilePrivateSerializer

from api.utils import get_is_following, get_is_follower, \
    get_is_follow_request_sent, get_count_of_followers, \
    get_count_of_follow_requests, get_count_of_followings, \
    get_is_follow_request_received, get_count_of_publications


class UserFullSerializer(serializers.ModelSerializer):
    """
    User serializer for admins.
    """
    profile = ProfileFullSerializer(many=True, read_only=True)
    is_follower = serializers.SerializerMethodField(read_only=True)
    is_following = serializers.SerializerMethodField(read_only=True)
    is_follow_request_sent = serializers.SerializerMethodField(read_only=True)
    is_follow_request_received = serializers. \
        SerializerMethodField(read_only=True)
    count_of_followers = serializers.SerializerMethodField(read_only=True)
    count_of_followings = serializers.SerializerMethodField(read_only=True)
    count_of_follow_requests = serializers. \
        SerializerMethodField(read_only=True)
    count_of_publications = serializers.SerializerMethodField(read_only=True)

    def get_is_follower(self, obj):
        request = self.context.get("request")
        return get_is_follower(request.user, obj)

    def get_is_following(self, obj):
        request = self.context.get("request")
        return get_is_following(request.user, obj)

    def get_is_follow_request_sent(self, obj):
        request = self.context.get("request")
        return get_is_follow_request_sent(request.user, obj)

    def get_is_follow_request_received(self, obj):
        request = self.context.get("request")
        return get_is_follow_request_received(request.user, obj)

    def get_count_of_followers(self, obj):
        return get_count_of_followers(obj)

    def get_count_of_followings(self, obj):
        return get_count_of_followings(obj)

    def get_count_of_follow_requests(self, obj):
        return get_count_of_follow_requests(obj)

    def get_count_of_publications(self, obj):
        return get_count_of_publications(obj)

    class Meta:
        model = User
        fields = ['id', 'username', 'email',
                  'profile', 'is_follower', 'is_following',
                  'is_follow_request_sent', 'is_active',
                  'is_follow_request_received',
                  'count_of_followers',
                  'count_of_followings',
                  'count_of_follow_requests',
                  'count_of_publications']


class UserBasicSerializer(serializers.ModelSerializer):
    """
    User serializer for users.
    """
    profile = ProfileBasicSerializer(many=True, read_only=True)
    is_follower = serializers.SerializerMethodField(read_only=True)
    is_following = serializers.SerializerMethodField(read_only=True)
    is_follow_request_sent = serializers.SerializerMethodField(read_only=True)
    is_follow_request_received = serializers. \
        SerializerMethodField(read_only=True)
    count_of_followers = serializers.SerializerMethodField(read_only=True)
    count_of_followings = serializers.SerializerMethodField(read_only=True)
    count_of_follow_requests = serializers. \
        SerializerMethodField(read_only=True)
    count_of_publications = serializers.SerializerMethodField(read_only=True)

    def get_is_follower(self, obj):
        request = self.context.get("request")
        return get_is_follower(request.user, obj)

    def get_is_following(self, obj):
        request = self.context.get("request")
        return get_is_following(request.user, obj)

    def get_is_follow_request_sent(self, obj):
        request = self.context.get("request")
        return get_is_follow_request_sent(request.user, obj)

    def get_is_follow_request_received(self, obj):
        request = self.context.get("request")
        return get_is_follow_request_received(request.user, obj)

    def get_count_of_followers(self, obj):
        return get_count_of_followers(obj)

    def get_count_of_followings(self, obj):
        return get_count_of_followings(obj)

    def get_count_of_follow_requests(self, obj):
        return get_count_of_follow_requests(obj)

    def get_count_of_publications(self, obj):
        return get_count_of_publications(obj)

    class Meta:
        model = User
        fields = ['id', 'username', 'profile', 'is_follower',
                  'is_following', 'is_follow_request_sent',
                  'is_follow_request_received',
                  'count_of_followers',
                  'count_of_followings',
                  'count_of_follow_requests',
                  'count_of_publications']


class UserPrivateSerializer(serializers.ModelSerializer):
    """
    User serializer for users.
    """
    profile = ProfilePrivateSerializer(many=True, read_only=True)
    is_follower = serializers.SerializerMethodField(read_only=True)
    is_following = serializers.SerializerMethodField(read_only=True)
    is_follow_request_sent = serializers.SerializerMethodField(read_only=True)
    is_follow_request_received = serializers. \
        SerializerMethodField(read_only=True)
    count_of_followers = serializers.SerializerMethodField(read_only=True)
    count_of_followings = serializers.SerializerMethodField(read_only=True)
    count_of_follow_requests = serializers. \
        SerializerMethodField(read_only=True)
    count_of_publications = serializers.SerializerMethodField(read_only=True)

    def get_is_follower(self, obj):
        request = self.context.get("request")
        return get_is_follower(request.user, obj)

    def get_is_following(self, obj):
        request = self.context.get("request")
        return get_is_following(request.user, obj)

    def get_is_follow_request_sent(self, obj):
        request = self.context.get("request")
        return get_is_follow_request_sent(request.user, obj)

    def get_is_follow_request_received(self, obj):
        request = self.context.get("request")
        return get_is_follow_request_received(request.user, obj)

    def get_count_of_followers(self, obj):
        return get_count_of_followers(obj)

    def get_count_of_followings(self, obj):
        return get_count_of_followings(obj)

    def get_count_of_follow_requests(self, obj):
        return get_count_of_follow_requests(obj)

    def get_count_of_publications(self, obj):
        return get_count_of_publications(obj)

    class Meta:
        model = User
        fields = ['id', 'username', 'profile', 'is_follower',
                  'is_following', 'is_follow_request_sent',
                  'is_follow_request_received',
                  'count_of_followers',
                  'count_of_followings',
                  'count_of_follow_requests',
                  'count_of_publications']


class UserNotificationSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ['id', 'username']



from api.models.collaboration_invite import CollaborationInvite
from rest_framework import viewsets
from rest_framework import permissions
from api.serializers.collaboration_invite \
    import CollaborationInviteSerializer, CollaborationInvitePOSTSerializer
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework import status
from api.permission import CollaborationPermissions
from django.shortcuts import get_object_or_404
from api.models.project import Project
from django.contrib.auth.models import User
from notifications.signals import notify
from django.utils import timezone
from rest_framework.exceptions import PermissionDenied


class CollaborationInviteViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = CollaborationInvite.objects.all()
    serializer_class = CollaborationInviteSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          CollaborationPermissions]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['from_user__id', 'to_user__id', 'to_project__id']

    def create(self, request, *args, **kwargs):

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.validated_data['created'] = timezone.now()
        serializer.validated_data['from_user'] = self.request.user
        project = Project.objects.get(
            id=serializer.validated_data['to_project'].id)
        to_user = User.objects.get(id=serializer.validated_data['to_user'].id)
        serializer.validated_data['to_user'] = to_user
        serializer.validated_data['to_project'] = project
        if (project.state == 'open for collaborators' or
            project.state == 'inviting collaborators') and \
                self.request.user.id != to_user.id:
            old_invite = CollaborationInvite.objects.filter(
                from_user=self.request.user,
                to_project=project,
                to_user=to_user)
            if old_invite.count() > 0:
                raise AlreadyInvitedException(
                    status_code=status.HTTP_400_BAD_REQUEST)
            col_invite = CollaborationInvite.objects.create(
                **serializer.validated_data)
            changed_data = {'id': col_invite.id}
            changed_data.update(serializer.data)

            ''' Invite Notification '''
            ''' Target --> Invite '''
            notify.send(sender=self.request.user,
                        verb="invited you to the Project {}"
                        .format(project.name),
                        recipient=to_user,
                        target=col_invite,
                        description="Invite"
                        )

            return Response(changed_data, status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    @action(detail=True, methods=['POST'], name='accept_collaboration_invite',
            serializer_class=None)
    def accept_collaboration_invite(self, request, pk=None):
        collaboration_invite = get_object_or_404(
            CollaborationInvite, pk=pk)
        if collaboration_invite.to_user == self.request.user:
            collaboration_invite.accept()

            ''' Accept Invite Notification '''
            ''' Target --> Project'''
            project = Project.objects.get(
                id=collaboration_invite.to_project_id)
            notify.send(sender=self.request.user,
                        verb="accepted your invitation to Project {}"
                        .format(project.name),
                        recipient=collaboration_invite.from_user,
                        target=project,
                        description="Project"
                        )
            return Response(status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    @action(detail=True, methods=['POST'], name='reject_collaboration_invite',
            serializer_class=None)
    def reject_collaboration(self, request, pk=None):

        collaboration_invite = get_object_or_404(
            CollaborationInvite, pk=pk)
        if collaboration_invite.to_user == self.request.user:
            collaboration_invite.reject()

            return Response(status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def get_serializer_class(self):
        if self.request.method == 'POST':
            return CollaborationInvitePOSTSerializer
        return super().get_serializer_class()


class AlreadyInvitedException(PermissionDenied):
    status_code = status.HTTP_400_BAD_REQUEST
    detail = "This user has already been invited."
    status_code = 'invalid'

    def __init__(self, status_code=None):
        if status_code is not None:
            self.status_code = status_code



# Generated by Django 3.1.4 on 2020-12-19 11:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0014_auto_20201217_1125'),
    ]

    operations = [
        migrations.AlterField(
            model_name='milestone',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='milestones', to='api.project'),
        ),
    ]



from rest_framework import serializers
from api.models.file import File


class FileSerializer(serializers.ModelSerializer):
    """
    File serializer
    """

    class Meta:
        model = File
        fields = ['id', 'file', 'remark', 'timestamp', 'project']



from api.models.notification import Notification
from rest_framework import serializers
from api.serializers.project import Project, ProjectPrivateSerializer
from api.serializers.profile import Profile, \
    ProfilePrivateNotificationSerializer
from api.serializers.user import User, UserNotificationSerializer
from api.serializers.collaboration_request import CollaborationRequest, \
    CollaborationRequestSerializer
from api.serializers.collaboration_invite import CollaborationInvite, \
    CollaborationInviteSerializer
from api.serializers.event import Event, EventSerializer
from api.serializers.file import File, FileSerializer
from api.serializers.milestone import Milestone, MilestoneSerializer
from api.serializers.tag import Tag, TagSerializer
from api.serializers.following import Following, FollowRequest, \
    FollowerSerializer, FollowRequestNotificationSerializer

from api.serializers.comment import Comment, CommentUpdateSerializer
from api.serializers.rating import Rating, RatingSerializer


class GenericNotificationRelatedField(serializers.RelatedField):
    def to_representation(self, value):
        if isinstance(value, Project):
            serializer = ProjectPrivateSerializer(value)
            return serializer.data
        if isinstance(value, Profile):
            serializer = ProfilePrivateNotificationSerializer(value)
        if isinstance(value, User):
            serializer = UserNotificationSerializer(value)
        if isinstance(value, CollaborationRequest):
            serializer = CollaborationRequestSerializer(value)
        if isinstance(value, CollaborationInvite):
            serializer = CollaborationInviteSerializer(value)
        if isinstance(value, Event):
            serializer = EventSerializer(value)
        if isinstance(value, File):
            serializer = FileSerializer(value)
        if isinstance(value, Milestone):
            serializer = MilestoneSerializer(value)
        if isinstance(value, Tag):
            serializer = TagSerializer(value)
        if isinstance(value, Following):
            serializer = FollowerSerializer(value)
        if isinstance(value, FollowRequest):
            serializer = FollowRequestNotificationSerializer(value)
        if isinstance(value, Comment):
            serializer = CommentUpdateSerializer(value)
        if isinstance(value, Rating):
            serializer = RatingSerializer(value)

        return serializer.data


class NotificationSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    target = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'target',
                  'unread', 'verb', 'timestamp']


class NotificationInviteSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    invite = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'invite',
                  'unread', 'verb', 'timestamp']


class NotificationRequestSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    request = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'request',
                  'unread', 'verb', 'timestamp']


class NotificationProjectSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    project = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'project',
                  'unread', 'verb', 'timestamp']


class NotificationFollowSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    following = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'following',
                  'unread', 'verb', 'timestamp']


class NotificationFollowRequestSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    follow_request = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'follow_request',
                  'unread', 'verb', 'timestamp']


class NotificationMilestoneSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    milestone = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'milestone',
                  'unread', 'verb', 'timestamp']


class NotificationCommentSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    comment = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'comment',
                  'unread', 'verb', 'timestamp']


class NotificationRatingSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    rating = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'rating',
                  'unread', 'verb', 'timestamp']


class NotificationUserSerializer(serializers.ModelSerializer):
    """
    Notification serializer
    """
    actor = GenericNotificationRelatedField(read_only=True)
    recipient = GenericNotificationRelatedField(read_only=True)
    user = GenericNotificationRelatedField(read_only=True)

    class Meta:
        model = Notification
        fields = ['id', 'actor', 'description',
                  'recipient', 'user',
                  'unread', 'verb', 'timestamp']



from rest_framework import serializers
from api.models.project import Project
from api.serializers.event import EventSerializer
from api.serializers.user import UserBasicSerializer
from api.serializers.tag import TagSerializer
from api.serializers.milestone import MilestoneSerializer


class ProjectPublicSerializer(serializers.ModelSerializer):
    """
    Project serializer, includes also profile owner
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')

    class Meta:
        model = Project
        fields = ['id', 'name', 'description', 'requirements', 'owner',
                  'is_public', 'state', 'project_type', 'due_date',
                  'event', 'tags', 'owner_id']


class ProjectGETPublicSerializer(serializers.ModelSerializer):
    """
    Project serializer, includes also profile owner
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')
    event = EventSerializer(read_only=True)
    members = UserBasicSerializer(many=True, read_only=True)
    tags = TagSerializer(many=True, read_only=True)
    milestones = MilestoneSerializer(many=True, read_only=True)

    class Meta:
        model = Project
        fields = ['id', 'name', 'description', 'requirements', 'owner',
                  'members', 'is_public', 'state', 'project_type', 'due_date',
                  'event', 'tags', 'milestones', 'owner_id']


class ProjectPrivateSerializer(serializers.ModelSerializer):
    """
    Project serializer, includes also profile owner
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')
    tags = TagSerializer(many=True, read_only=True)

    class Meta:
        model = Project
        fields = ['id', 'name', 'description', 'owner',
                  'project_type', 'tags', 'is_public',
                  'state', 'owner_id']


class ProjectMilestonesSerializer(serializers.ModelSerializer):
    """
    Project serializer, includes also profile owner
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')
    milestones = MilestoneSerializer(many=True, read_only=True)

    class Meta:
        model = Project
        fields = ['id', 'name', 'description', 'owner',
                  'project_type', 'milestones', 'is_public',
                  'state', 'owner_id']



from django.db import models


class Publication(models.Model):
    """
    Publication model
    """
    title = models.CharField(max_length=1000)
    publication_year = models.IntegerField(default=0)
    link = models.TextField(default='')
    citation_number = models.IntegerField(default=0)
    owner = models.ForeignKey(
        'auth.User', related_name='publication_owner',
        on_delete=models.CASCADE)

    class Meta:
        ordering = ['citation_number']



# Generated by Django 3.1.4 on 2021-01-17 10:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0018_merge_20210107_2115'),
    ]

    operations = [
        migrations.AlterField(
            model_name='file',
            name='remark',
            field=models.CharField(max_length=100),
        ),
    ]



from rest_framework import viewsets
from rest_framework import permissions
from api.models.file import File
from api.serializers.file import FileSerializer
from api.permission import IsFileMember
from rest_framework.parsers import MultiPartParser, FormParser
from rest_framework.decorators import action
from django.http import FileResponse
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.response import Response
from django.db.models import Q


class FileViewSet(viewsets.ModelViewSet):

    parser_classes = (MultiPartParser, FormParser)
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = File.objects.all()
    serializer_class = FileSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          IsFileMember]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['project']

    '''
    Custom list action for file.
    Non-admin users will not be able to see files of a project that is
    private and they are not the owner or a member.
    '''

    def list(self, request, *args, **kwargs):
        user = request.user
        queryset = self.filter_queryset(self.get_queryset())

        if not user.is_staff:
            queryset = queryset.filter(Q(project__is_public__exact=True) |
                                       Q(project__owner__exact=user.id) |
                                       Q(project__members__exact=user.id))

        queryset = queryset.distinct()
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data,
                                         partial=partial)
        serializer.is_valid(raise_exception=True)
        if instance.file:
            instance.file.delete()
        self.perform_update(serializer)
        return Response(serializer.data)

    @action(detail=True)
    def retrieve_file(self, request, pk=None):
        filename = self.get_object().file
        return FileResponse(filename)

    @action(detail=True)
    def download_file(self, request, pk=None):
        filename = self.get_object().file
        return FileResponse(filename, as_attachment=True)



"""
Django settings for backend project.
Generated by 'django-admin startproject' using Django 3.1.3.
For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/
For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

# comment to check pylint

from pathlib import Path
import django_heroku
import os

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '+@j*1ow05x29cj^-j789-z3kntfu+w17sm*ngs-xxbgp)%27m1'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['*']

# Setup support for proxy headers
USE_X_FORWARDED_HOST = True
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'corsheaders',
    'django_filters',
    'api',
    'rest_framework',
    'rest_framework.authtoken',
    'drf_yasg',
    'password_reset',
    'notifications',
    'storages',
    'django_email_verification'
]

MIDDLEWARE = [
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'backend.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR / 'templates'],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.TokenAuthentication',
    ),
    'TEST_REQUEST_DEFAULT_FORMAT': 'json'
}

DJANGO_NOTIFICATIONS_CONFIG = {'USE_JSONFIELD': True}
NOTIFICATIONS_NOTIFICATION_MODEL = 'api.Notification'

WSGI_APPLICATION = 'backend.wsgi.application'

# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

if os.environ.get('DJANGO_ENV', 'development') != 'production':
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        }
    }
else:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': os.environ.get('DB_NAME', 'paperlayer'),
            'USER': os.environ.get('DB_USER', 'paperlayer'),
            'PASSWORD': os.environ.get('DB_PASSWORD', 'paperlayer'),
            'HOST': os.environ.get('DB_ADDRESS', '127.0.0.1'),
            'PORT': os.environ.get('DB_PORT', 5432)
        }
    }

SWAGGER_SETTINGS = {
    'SECURITY_DEFINITIONS': {
        'Token': {
            'type': 'apiKey',
            'name': 'Authorization',
            'in': 'header',
            'description': 'Please write "Token" before the key',
        }
    },
    'USE_SESSION_AUTH': False,
}

# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.'
                'UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.'
                'MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.'
                'CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.'
                'NumericPasswordValidator',
    },
]

# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = '/static/'

# SMTP Configuration

EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
EMAIL_HOST = 'smtp.gmail.com'
EMAIL_PORT = 587
EMAIL_USE_TLS = True
EMAIL_HOST_USER = 'bounswe2020group3@gmail.com'
EMAIL_HOST_PASSWORD = os.environ.get('EMAIL_HOST_PASSWORD', '')

EMAIL_ACTIVE_FIELD = 'is_active'
EMAIL_PAGE_TEMPLATE = 'verification_confirm.html'

django_heroku.settings(locals())

# Cors settings
CORS_ORIGIN_ALLOW_ALL = True

# AWS S3 for files
AWS_ACCESS_KEY_ID = os.environ.get('AWS_ACCESS_KEY_ID', '')
AWS_SECRET_ACCESS_KEY = os.environ.get('AWS_SECRET_ACCESS_KEY', '')
DEFAULT_FILE_STORAGE = 'storages.backends.s3boto3.S3Boto3Storage'
AWS_STORAGE_BUCKET_NAME = 'paperlayerbucket'
AWS_S3_REGION_NAME = 'eu-central-1'
AWS_S3_CUSTOM_DOMAIN = '%s.s3.amazonaws.com' % AWS_STORAGE_BUCKET_NAME

STREAM_API_KEY = os.environ.get('STREAM_API_KEY', '')
STREAM_API_SECRET = os.environ.get('STREAM_API_SECRET', '')



from api.models.rating import Rating
from api.models.profile import Profile
from api.models.following import Following
from django.contrib.auth.models import User
from rest_framework import viewsets
from rest_framework import permissions
from api.permission import RatingPermission
from api.serializers.rating import RatingSerializer, \
    RatingUpdateSerializer, RatingFeedSerializer
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.response import Response
from notifications.signals import notify
from rest_framework import status

from api.views.feed import add_activity_to_feed


class RatingViewSet(viewsets.ModelViewSet):
    """
    Users can read, update, partial_update or delete only the
    ratings that themselves created.
    """
    queryset = Rating.objects.all()
    serializer_class = RatingSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          RatingPermission]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['to_user']

    def get_serializer_class(self):
        if self.action in ['update', 'partial_update']:
            return RatingUpdateSerializer
        else:
            return RatingSerializer

    def perform_create(self, serializer):
        serializer.save(from_user=self.request.user)

    def create(self, request, *args, **kwargs):
        """
        Rating can be an integer in [0,10].
        The from_user field is automatically the requesting user.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.validated_data['from_user'] = self.request.user
        to_user = User.objects.get(
            id=serializer.validated_data['to_user'].id)
        serializer.validated_data['to_user'] = to_user

        to_profile = Profile.objects.get(owner=to_user)
        is_public = to_profile.is_public

        is_following = Following.objects. \
            filter(from_user=self.request.user,
                   to_user=to_user).exists()

        if (is_public or is_following) and \
                self.request.user.id != to_user.id:
            rating = Rating.objects.create(**serializer.validated_data)

            """
                Rating Notification
            """

            notify.send(sender=self.request.user,
                        verb="rated about you",
                        recipient=to_user,
                        target=rating,
                        description="Rating"
                        )

            """
                Adds the rating to user feed.
            """

            rating_serializer = RatingFeedSerializer(rating).data
            activity_data = {'actor': str(self.request.user),
                             'verb': 'rated {} {} stars'.format(
                                 rating_serializer['to_user']['username'],
                                 rating_serializer['rating']),
                             'type': 'rating',
                             'object': rating_serializer['id'],
                             'foreign_id': 'rating:' +
                                           str(rating_serializer['id']),
                             'rating': rating_serializer,
                             }

            add_activity_to_feed(self.request.user, activity_data)

            return Response(serializer.data, status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def list(self, request, *args, **kwargs):
        """
        Users can only see the ratings they created.
        """
        user = request.user
        queryset = self.filter_queryset(self.get_queryset())
        if user.is_anonymous:
            queryset = []
        elif not user.is_staff:
            queryset = queryset. \
                       filter(from_user__exact=user.id)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)



from rest_framework import serializers
from api.models.publication import Publication


class PublicationSerializer(serializers.ModelSerializer):
    """
    Publication serializer
    """

    class Meta:
        model = Publication
        fields = '__all__'



from api.models.report import Report
from rest_framework import permissions
from rest_framework import viewsets
from api.serializers.report import ReportSerializer
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.response import Response
from rest_framework import status


class ReportViewSet(viewsets.GenericViewSet):
    queryset = Report.objects.all()
    serializer_class = ReportSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['reported_user']

    def retrieve(self, request, *args, **kwargs):
        '''
        Only admins can view reports.
        '''
        user = request.user
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        if user.is_staff:
            return Response(serializer.data)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def list(self, request, *args, **kwargs):
        '''
        Only admins can view reports.
        '''
        user = request.user
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None and user.is_staff:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)

        if user.is_staff:
            return Response(serializer.data)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def destroy(self, request, *args, **kwargs):
        '''
        Only admins can delete reports.
        '''
        user = request.user
        if user.is_staff:
            instance = self.get_object()
            self.perform_destroy(instance)
            return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def create(self, request, *args, **kwargs):
        '''
        Only signed-in users can create reports.
        REPORT_TYPES = ["Disturbing other users",
                "Sharing unrelated or disturbing posts",
                "Spam", "Fake Profile", "Stolen Account"]
        '''
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        return Response(serializer.data,
                        status=status.HTTP_201_CREATED)

    def perform_create(self, serializer):
        serializer.save()

    def perform_destroy(self, instance):
        instance.delete()



# Generated by Django 3.1.3 on 2020-12-14 17:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0010_merge_20201214_2048'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='profile',
            name='photo_url',
        ),
        migrations.AddField(
            model_name='profile',
            name='affiliations',
            field=models.TextField(blank=True, default=''),
        ),
        migrations.AddField(
            model_name='profile',
            name='profile_picture',
            field=models.ImageField(blank=True, null=True, upload_to=''),
        ),
    ]



# Generated by Django 3.1.4 on 2021-01-17 14:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0016_auto_20210117_1657'),
        ('api', '0019_auto_20210117_1056'),
    ]

    operations = [
    ]



from api.models.following import Following, FollowRequest
from api.models.project import Project
from api.models.rating import Rating
from api.models.publication import Publication
from django.db.models import Avg


def get_is_following(this_user, accessed_user):
    """
        Returns True if "this_user" follows "accessed_user"
    """
    if not this_user or this_user.is_anonymous:
        is_following = False
    else:
        is_following = \
            Following.objects.filter(
                from_user=this_user,
                to_user=accessed_user).exists()
    return is_following


def get_is_follower(this_user, accessed_user):
    """
        Returns True if "accessed_user" follows "this_user"
    """
    if not this_user or this_user.is_anonymous:
        is_follower = False
    else:
        is_follower = \
            Following.objects.filter(
                from_user=accessed_user,
                to_user=this_user).exists()
    return is_follower


def get_is_follow_request_sent(this_user, accessed_user):
    """
        Returns True if "this_user" has
        sent follow request to "accessed_user"
    """
    if not this_user or this_user.is_anonymous:
        is_is_follow_request_sent = False
    else:
        is_is_follow_request_sent = \
            FollowRequest.objects.filter(
                req_from_user=this_user,
                req_to_user=accessed_user).exists()
    return is_is_follow_request_sent


def get_is_follow_request_received(this_user, accessed_user):
    """
        Returns True if "this_user" has
        gotten follow request from "accessed_user"
    """
    if not this_user or this_user.is_anonymous:
        is_is_follow_request_received = False
    else:
        is_is_follow_request_received = \
            FollowRequest.objects.filter(
                req_from_user=accessed_user,
                req_to_user=this_user).exists()
    return is_is_follow_request_received


def get_is_collaborator(this_user, accessed_user):
    """
    Returns True if "this_user" and "accessed_user" has a project that
    they are both a member.
    """
    if not this_user or this_user.is_anonymous:
        is_collaborator = False
    else:
        common = Project.objects. \
            filter(members__exact=this_user.id)
        is_collaborator = common. \
            filter(members__exact=accessed_user.id).exists()

    return is_collaborator


def get_my_rating(this_user, accessed_user):
    if this_user.is_anonymous:
        return "You need to log in to see this."
    rating = Rating.objects.filter(to_user__exact=accessed_user,
                                   from_user__exact=this_user)
    is_rating_exists = rating.exists()
    if is_rating_exists:
        return rating[0].rating
    else:
        return "You did not rate this user."


def get_my_rating_id(this_user, accessed_user):
    if this_user.is_anonymous:
        return "You need to log in to see this."
    rating = Rating.objects.filter(to_user__exact=accessed_user,
                                   from_user__exact=this_user)
    is_rating_exists = rating.exists()
    if is_rating_exists:
        return rating[0].id
    else:
        return "You did not rate this user."


def get_count_of_followers(this_user):
    """
        Returns length of followers of user
    """
    return Following.objects.filter(to_user=this_user.id).count()


def get_count_of_followings(this_user):
    """
        Returns length of followings of user
    """
    return Following.objects.filter(from_user=this_user.id).count()


def get_count_of_follow_requests(this_user):
    """
        Returns length of follow requests coming to user.
    """
    return FollowRequest.objects.filter(req_to_user=this_user.id).count()


def get_count_of_publications(this_user):
    """
        Returns length of publications of user
    """
    return Publication.objects.filter(owner=this_user.id).count()


def get_user_rating(user_id):
    ratings = Rating.objects.filter(to_user=user_id)
    rating = ratings.aggregate(Avg('rating'))
    return rating['rating__avg']



from rest_framework import viewsets
from rest_framework import permissions
from api.models.milestone import Milestone
from api.models.project import Project
from api.serializers.project import ProjectMilestonesSerializer
from api.serializers.milestone import MilestoneSerializer
from api.permission import IsMilestoneMemberOrReadOnly
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework import status
from django_filters.rest_framework import DjangoFilterBackend
from notifications.signals import notify
from django.contrib.auth.models import User

from api.serializers.user import UserNotificationSerializer


class MilestoneViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = Milestone.objects.all()
    serializer_class = MilestoneSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          IsMilestoneMemberOrReadOnly]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['project__id']

    @action(detail=False, methods=['GET'], url_name='get_user_milestones')
    def get_user_milestones(self, request, *args):
        username = self.request.user.username
        list_of_milestones = []
        if username:
            projects = Project.objects.filter(members__username__in=[username])
            serializer = ProjectMilestonesSerializer(projects, many=True)
            for item in serializer.data:
                for milestone in item['milestones']:
                    milestone['project_name'] = item['name']
                    list_of_milestones.append(milestone)
            list_of_milestones.sort(key=lambda x: x.get('date'))
            return Response(data={
                'result': list_of_milestones,
            }, status=status.HTTP_200_OK)
        else:
            return Response(data={
                'result': 'Not found'
            }, status=status.HTTP_404_NOT_FOUND)

    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)
        if self.action == 'create':
            milestone = Milestone.objects.get(id=response.data['id'])
            members = UserNotificationSerializer(milestone.project.members,
                                                 many=True)
            for member in members.data:
                user = User.objects.get(id=member['id'])
                notify.send(sender=self.request.user,
                            verb="created a new Milestone with description {}".
                            format(response.data['description']),
                            recipient=user,
                            target=milestone,
                            description='Milestone')
        if self.action == 'update':
            pass
        if self.action == 'destroy':
            pass
        return response



from django.contrib.auth.models import User
from rest_framework import mixins, viewsets, status

import stream
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework import serializers

from backend import settings


def client():
    return stream.connect(settings.STREAM_API_KEY, settings.STREAM_API_SECRET)


def add_activity_to_feed(username, activity_data):
    user = User.objects.get(username=username)
    user_feed = client().feed('user', user.id)
    user_feed.add_activity(activity_data)


def follow_admin(username):
    user = User.objects.get(username=username)
    user_feed = client().feed('timeline', user.id)
    admin_id = User.objects.get(username='admin').id
    user_feed.follow('user', admin_id)


def follow_user_timeline(username, followed_username):
    user_id = User.objects.get(username=username).id
    followed_username_id = User.objects.get(username=followed_username).id
    user_feed = client().feed('timeline', user_id)
    user_feed.follow('user', followed_username_id)


class FeedSerializer(serializers.Serializer):
    user_id = serializers.IntegerField()


class FeedViewSet(mixins.ListModelMixin, mixins.DestroyModelMixin,
                  viewsets.GenericViewSet):
    queryset = None
    permission_classes = [IsAuthenticated]

    def list(self, request, *args, **kwargs):
        user_id = User.objects.get(username=self.request.user).id
        user_feed = client().feed('user', user_id)
        return Response(data={
            'results': user_feed.get()['results']
        })

    def destroy(self, request, *args, **kwargs):
        user_id = User.objects.get(username=self.request.user).id
        user_feed = client().feed('user', user_id)
        response = user_feed.remove_activity(self.kwargs['pk'])
        return Response(data=response, status=status.HTTP_204_NO_CONTENT)

    @action(detail=False, methods=['GET'], url_name='timeline',
            serializer_class=FeedSerializer)
    def timeline(self, request):
        user_feed = self.get_timeline()
        results = user_feed.get()['results']
        return Response(data={
            'count': len(results),
            'results': results
        }, status=status.HTTP_200_OK)

    @action(detail=False, methods=['POST'], url_name='unfollow_timeline',
            serializer_class=FeedSerializer)
    def unfollow(self, request):
        unfollow_user_id = self.request.data['user_id']
        user_feed = self.get_timeline()
        response = user_feed.unfollow('user', unfollow_user_id)
        return Response(data=response, status=status.HTTP_200_OK)

    @action(detail=False, methods=['GET'], url_name='timeline_followers')
    def followers(self, request):
        user_feed = self.get_timeline()
        response = user_feed.following()
        return Response(data=response, status=status.HTTP_200_OK)

    def get_timeline(self):
        user = User.objects.get(username=self.request.user)
        return client().feed('timeline', user.id)



# Generated by Django 3.1.3 on 2020-11-22 18:15

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100)),
                ('description', models.CharField(blank=True, default='', max_length=200)),
                ('deadline', models.DateField()),
                ('date', models.DateField()),
                ('event_type', models.CharField(choices=[('journal submission', '0'), ('academic conference', '1'), ('funded project', '2')], max_length=50)),
                ('url', models.URLField(blank=True, default='')),
            ],
            options={
                'ordering': ['date'],
            },
        ),
        migrations.CreateModel(
            name='Example',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('description', models.CharField(max_length=255)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=500)),
                ('description', models.TextField(default='')),
                ('requirements', models.TextField(default='')),
                ('is_public', models.BooleanField(default=False)),
                ('state', models.CharField(choices=[('seeking for collaborators', '0'), ('open for collaborators', '1'), ('in progress', '2'), ('done', '3')], default='seeking for collaborators', max_length=100)),
                ('project_type', models.CharField(choices=[('conference', '0'), ('instutution', '1'), ('journal', '2')], default='conference', max_length=100)),
                ('due_date', models.DateField(default=datetime.date.today)),
                ('events', models.ManyToManyField(blank=True, to='api.Event')),
                ('members', models.ManyToManyField(related_name='members', to=settings.AUTH_USER_MODEL)),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='owner', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('projects', models.ManyToManyField(to='api.Project')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(default='', max_length=100)),
                ('middle_name', models.CharField(blank=True, default='', max_length=100)),
                ('last_name', models.CharField(default='', max_length=100)),
                ('bio', models.CharField(blank=True, default='', max_length=1000)),
                ('photo_url', models.TextField(blank=True, default='')),
                ('age', models.IntegerField(default=0)),
                ('share_age', models.BooleanField(default=True)),
                ('expertise', models.TextField(blank=True, default='')),
                ('gender', models.CharField(choices=[('male', '0'), ('female', '1'), ('do not want to share', '2')], default='do not want to share', max_length=100)),
                ('interests', models.TextField(blank=True, default='')),
                ('share_bio', models.BooleanField(default=True)),
                ('share_gender', models.BooleanField(default=True)),
                ('share_affiliations', models.BooleanField(default=True)),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.CreateModel(
            name='Milestone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(default='')),
                ('date', models.DateField()),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.project')),
            ],
            options={
                'ordering': ['project'],
            },
        ),
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', models.FileField(upload_to='')),
                ('remark', models.CharField(max_length=20)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.project')),
            ],
            options={
                'ordering': ['project'],
            },
        ),
    ]



# Generated by Django 3.1.3 on 2020-12-06 20:24

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0004_merge_20201206_1815'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='members',
            field=models.ManyToManyField(null=True, related_name='members', to=settings.AUTH_USER_MODEL),
        ),
    ]



from api.models.comment import Comment
from api.models.profile import Profile
from rest_framework import serializers

from api.serializers.user import UserNotificationSerializer


class CommentSerializer(serializers.ModelSerializer):
    """
    Comment serializer
    """
    created = serializers.DateTimeField(read_only=True)
    owner = serializers.ReadOnlyField(source='from_user.username')
    name = serializers.SerializerMethodField('get_name')
    middle_name = serializers.SerializerMethodField('get_middle_name')
    last_name = serializers.SerializerMethodField('get_last_name')

    class Meta:
        model = Comment
        fields = ['id', 'created', 'comment', 'from_user', 'to_user',
                  'owner', 'name', 'middle_name', 'last_name']

    def get_name(self, obj):
        profile = Profile.objects.filter(owner__exact=obj.from_user)
        is_profile_exists = profile.exists()
        if is_profile_exists:
            return profile[0].name
        else:
            return "There is no such profile."

    def get_middle_name(self, obj):
        profile = Profile.objects.filter(owner__exact=obj.from_user)
        is_profile_exists = profile.exists()
        if is_profile_exists:
            return profile[0].middle_name
        else:
            return "There is no such profile."

    def get_last_name(self, obj):
        profile = Profile.objects.filter(owner__exact=obj.from_user)
        is_profile_exists = profile.exists()
        if is_profile_exists:
            return profile[0].last_name
        else:
            return "There is no such profile."


class CommentUpdateSerializer(serializers.ModelSerializer):
    """
    Comment update serializer
    """
    class Meta:
        model = Comment
        fields = ['id', 'comment']


class CommentCreateSerializer(serializers.ModelSerializer):
    """
    Comment serializer
    """
    created = serializers.DateTimeField(read_only=True)
    owner = serializers.ReadOnlyField(source='from_user.username')

    class Meta:
        model = Comment
        fields = ['id', 'created', 'comment', 'from_user', 'to_user',
                  'owner']


class CommentFeedSerializer(serializers.ModelSerializer):
    """
    Comment update serializer
    """

    from_user = UserNotificationSerializer(read_only=True)
    to_user = UserNotificationSerializer(read_only=True)

    class Meta:
        model = Comment
        fields = ['id', 'comment', 'from_user', 'to_user']



from django.contrib.auth.models import User
from api.models.profile import Profile
from rest_framework import serializers
from django.utils.translation import gettext_lazy as _


class RegisterSerializer(serializers.Serializer):
    first_name = serializers.CharField()
    middle_name = serializers.CharField(allow_blank=True)
    last_name = serializers.CharField()

    username = serializers.CharField()
    email = serializers.EmailField()
    password = serializers.CharField()

    def save(self, **kwargs):
        data = self.validated_data
        user = User.objects.create_user(data['username'], email=data['email'],
                                        password=data['password'],
                                        is_active=False)
        user.save()
        profile = Profile(owner=user, name=data['first_name'],
                          middle_name=data['middle_name'],
                          last_name=data['last_name'])
        profile.save()
        return user


class AuthTokenSerializer(serializers.Serializer):
    username = serializers.CharField(
        label=_("Username"),
        write_only=True
    )
    password = serializers.CharField(
        label=_("Password"),
        style={'input_type': 'password'},
        trim_whitespace=False,
        write_only=True
    )
    token = serializers.CharField(
        label=_("Token"),
        read_only=True
    )

    def authenticate_email(self, email: str, password: str):
        """ Authenticate a user based on username. """
        try:
            user = User.objects.get(email=email)
            if user.check_password(password):
                return user
        except Exception:
            return None

    def authenticate_username(self, username: str, password: str):
        """ Authenticate a user based on email address. """
        try:
            user = User.objects.get(username=username)
            if user.check_password(password):
                return user
        except Exception:
            return None

    def validate(self, attrs):
        username = attrs.get('username')
        password = attrs.get('password')

        if username and password:
            user = self.authenticate_username(
                username=username, password=password)
            if not user:
                user = self.authenticate_email(
                    email=username, password=password)

            # The authenticate call simply returns None for is_active=False
            # users. (Assuming the default ModelBackend authentication
            # backend.)
            if not user:
                msg = _('Unable to log in with provided credentials.')
                raise serializers.ValidationError(msg, code='authorization')
        else:
            msg = _('Must include "username" and "password".')
            raise serializers.ValidationError(msg, code='authorization')

        attrs['user'] = user
        return attrs



from django.contrib.auth.models import User
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from api.serializers import auth
from django.contrib.auth import logout
from rest_framework import generics
from rest_framework import parsers, renderers
from rest_framework.authtoken.models import Token
from api.serializers.auth import AuthTokenSerializer
from rest_framework.compat import coreapi, coreschema
from rest_framework.schemas import ManualSchema
from rest_framework.schemas import coreapi as coreapi_schema
from rest_framework.views import APIView

# from django.core.mail import send_mail
from base64 import urlsafe_b64encode
from django.template.loader import render_to_string
from django_email_verification.views import verify
from django.urls import get_resolver
from django.contrib.auth.tokens import default_token_generator
import sendgrid
from sendgrid.helpers.mail import Email, To, Content, Mail
import os

from api.views.feed import follow_admin


class RegisterGenericAPIView(generics.GenericAPIView):
    serializer_class = auth.RegisterSerializer

    def __send_verification(self, user: User):
        domain = 'https://paperlayer-backend.azurewebsites.net/'
        token = default_token_generator.make_token(user)
        email = urlsafe_b64encode(str(user.email).encode('utf-8'))
        token = f'{email.decode("utf-8")}/{token}'

        link = ''
        for k, v in get_resolver(None).reverse_dict.items():
            if k is verify and v[0][0][1][0]:
                addr = str(v[0][0][0])
                link = domain + addr[0: addr.index('%')] + token

        text = render_to_string('verification_body.txt', {'link': link})

        sg = sendgrid.SendGridAPIClient(
            api_key=os.environ.get('SENDGRID_API_KEY'))
        from_email = Email("bounswe2020group3@gmail.com")
        to_email = To(user.email)
        subject = "Confirm your PaperLayer account"
        content = Content("text/plain", text)
        mail = Mail(from_email, to_email, subject, content)
        sg.client.mail.send.post(request_body=mail.get())

    def post(self, request, *args, **kwargs):
        """
        Registers a new user
        """
        serializer = auth.RegisterSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.save()
        self.__send_verification(user)

        follow_admin(serializer.data['username'])

        return Response(serializer.data, status=status.HTTP_201_CREATED)


class LogoutGenericAPIView(generics.GenericAPIView):
    permission_classes = [IsAuthenticated]

    def post(self, request, *args, **kwargs):
        """
        Logs the user out, has to be authenticated
        """
        logout(request)
        return Response(None, status=status.HTTP_200_OK)


class AuthView(APIView):
    throttle_classes = ()
    permission_classes = ()
    parser_classes = (parsers.FormParser,
                      parsers.MultiPartParser, parsers.JSONParser,)
    renderer_classes = (renderers.JSONRenderer,)
    serializer_class = AuthTokenSerializer

    if coreapi_schema.is_enabled():
        schema = ManualSchema(
            fields=[
                coreapi.Field(
                    name="username",
                    required=True,
                    location='form',
                    schema=coreschema.String(
                        title="Username",
                        description="Valid username for authentication",
                    ),
                ),
                coreapi.Field(
                    name="password",
                    required=True,
                    location='form',
                    schema=coreschema.String(
                        title="Password",
                        description="Valid password for authentication",
                    ),
                ),
            ],
            encoding="application/json",
        )

    def get_serializer_context(self):
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }

    def get_serializer(self, *args, **kwargs):
        kwargs['context'] = self.get_serializer_context()
        return self.serializer_class(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data['user']
        token, created = Token.objects.get_or_create(user=user)
        return Response({'id': user.pk, 'token': token.key})



from api.models.profile import Profile
from rest_framework import serializers
from api.models.rating import Rating
from django.db.models import Avg
from api.utils import get_my_rating, get_my_rating_id, \
    get_is_collaborator


class ProfileFullSerializer(serializers.HyperlinkedModelSerializer):
    """
    Profile serializer for owner and admin.
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')
    email = serializers.ReadOnlyField(source='owner.email')
    rating = serializers.SerializerMethodField('get_rating')
    my_rating = serializers.SerializerMethodField('get_my_rating')
    my_rating_id = serializers.SerializerMethodField('get_my_rating_id')
    is_commentable = serializers.SerializerMethodField('get_is_commentable')

    class Meta:
        model = Profile
        fields = ['id', 'name', 'middle_name', 'last_name', 'owner', 'email',
                  'bio', 'profile_picture', 'birthday', 'share_birthday',
                  'expertise', 'gender', 'interests', 'affiliations',
                  'share_bio', 'share_gender', 'share_affiliations',
                  'is_public', 'rating', 'my_rating', 'my_rating_id',
                  'is_commentable', 'owner_id']

    def get_rating(self, obj):
        ratings = Rating.objects.filter(to_user__exact=obj.owner)
        rating = ratings.aggregate(Avg('rating'))
        return rating['rating__avg']

    def get_my_rating(self, obj):
        request = self.context.get("request")
        return get_my_rating(request.user, obj.owner)

    def get_my_rating_id(self, obj):
        request = self.context.get("request")
        return get_my_rating_id(request.user, obj.owner)

    def get_is_commentable(self, obj):
        request = self.context.get("request")
        return get_is_collaborator(request.user, obj.owner)


class ProfileBasicSerializer(serializers.HyperlinkedModelSerializer):
    """
    Public profile serializer for others.
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')
    email = serializers.ReadOnlyField(source='owner.email')
    rating = serializers.SerializerMethodField('get_rating')
    my_rating = serializers.SerializerMethodField('get_my_rating')
    my_rating_id = serializers.SerializerMethodField('get_my_rating_id')
    is_commentable = serializers.SerializerMethodField('get_is_commentable')

    class Meta:
        model = Profile
        fields = ['id', 'name', 'middle_name', 'last_name', 'owner', 'email',
                  'bio', 'profile_picture', 'birthday', 'share_birthday',
                  'expertise',
                  'gender', 'interests', 'affiliations', 'share_bio',
                  'share_gender', 'share_affiliations',
                  'is_public', 'rating', 'my_rating', 'my_rating_id',
                  'is_commentable', 'owner_id']

    def get_rating(self, obj):
        ratings = Rating.objects.filter(to_user__exact=obj.owner)
        rating = ratings.aggregate(Avg('rating'))
        return rating['rating__avg']

    def get_my_rating(self, obj):
        request = self.context.get("request")
        return get_my_rating(request.user, obj.owner)

    def get_my_rating_id(self, obj):
        request = self.context.get("request")
        return get_my_rating_id(request.user, obj.owner)

    def get_is_commentable(self, obj):
        request = self.context.get("request")
        return get_is_collaborator(request.user, obj.owner)

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        if not ret['share_birthday']:
            ret['birthday'] = ""

        if not ret['share_gender']:
            ret['gender'] = ""

        if not ret['share_bio']:
            ret['bio'] = ""

        if not ret['share_affiliations']:
            ret['affiliations'] = ""

        return ret


class ProfilePrivateSerializer(serializers.HyperlinkedModelSerializer):
    """
    Private profile serializer.
    """
    owner = serializers.ReadOnlyField(source='owner.username')
    owner_id = serializers.ReadOnlyField(source='owner.id')
    my_rating = serializers.SerializerMethodField('get_my_rating')
    my_rating_id = serializers.SerializerMethodField('get_my_rating_id')
    is_commentable = serializers.SerializerMethodField('get_is_commentable')

    class Meta:
        model = Profile
        fields = ['id', 'name', 'middle_name', 'last_name',
                  'owner', 'profile_picture', 'is_public', 'my_rating',
                  'my_rating_id', 'is_commentable', 'owner_id']

    def get_my_rating(self, obj):
        request = self.context.get("request")
        return get_my_rating(request.user, obj.owner)

    def get_my_rating_id(self, obj):
        request = self.context.get("request")
        return get_my_rating_id(request.user, obj.owner)

    def get_is_commentable(self, obj):
        request = self.context.get("request")
        return get_is_collaborator(request.user, obj.owner)


class ProfilePictureSerializer(serializers.HyperlinkedModelSerializer):
    """
    Profile serializer for owner and admin.
    """

    class Meta:
        model = Profile
        fields = ['id', 'profile_picture']


class ProfilePrivateNotificationSerializer(serializers.
                                           HyperlinkedModelSerializer):
    """
    Private profile serializer.
    """
    owner = serializers.ReadOnlyField(source='owner.username')

    class Meta:
        model = Profile
        fields = ['id', 'name', 'middle_name', 'last_name',
                  'owner', 'is_public']



"""backend URL Configuration
The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.urls import path, include
from django.conf.urls import url
from .swagger_api import schema_view
from rest_framework.routers import DefaultRouter

router = DefaultRouter()

urlpatterns = [
    url(r'^swagger(?P<format>\.json|\.yaml)$',
        schema_view.without_ui(cache_timeout=0), name='schema-json'),
    url(r'^swagger/$', schema_view.with_ui('swagger',
                                           cache_timeout=0),
        name='schema-swagger-ui'),
    url(r'^redoc/$', schema_view.with_ui('redoc',
                                         cache_timeout=0),
        name='schema-redoc'),
    path('admin/', admin.site.urls),
    path('api/', include('api.urls')),
    path('', include(router.urls)),
]



from django.db import models


class Following(models.Model):
    """
    Following model for follows.
    """
    from_user = models.ForeignKey('auth.User',
                                  related_name="following",
                                  on_delete=models.CASCADE)
    to_user = models.ForeignKey('auth.User',
                                related_name="followers",
                                on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['from_user',
                                            'to_user'],
                                    name="unique_followers")
        ]
        ordering = ['created']

    def unfollow(self):
        self.delete()
        Following.objects.filter(
            from_user=self.to_user,
            to_user=self.from_user
        ).delete()


class FollowRequest(models.Model):
    req_from_user = models.ForeignKey('auth.User',
                                      related_name="sent_requests",
                                      on_delete=models.CASCADE)
    req_to_user = models.ForeignKey('auth.User',
                                    related_name="follow_requests",
                                    on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['req_from_user',
                                            'req_to_user'],
                                    name="unique_request")
        ]
        ordering = ['created']

    def accept(self):
        Following.objects.create(from_user=self.req_from_user,
                                 to_user=self.req_to_user).save()
        self.delete()
        FollowRequest.objects.filter(
            req_from_user=self.req_to_user,
            req_to_user=self.req_from_user
        ).delete()
        # TODO add notification signal

    def reject(self):
        self.delete()
        FollowRequest.objects.filter(
            req_from_user=self.req_to_user,
            req_to_user=self.req_from_user
        ).delete()
        # TODO add notification signal2



# Generated by Django 3.1.3 on 2020-12-09 11:36

import api.models.tag
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0008_auto_20201206_2028'),
    ]

    operations = [
        migrations.RenameField(
            model_name='profile',
            old_name='share_age',
            new_name='share_birthday',
        ),
        migrations.RemoveField(
            model_name='profile',
            name='age',
        ),
        migrations.AddField(
            model_name='profile',
            name='birthday',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='tag',
            name='color',
            field=models.IntegerField(default=api.models.tag.random_color),
        ),
    ]



from rest_framework import serializers
from api.serializers.event import EventSerializer
from api.serializers.project import ProjectGETPublicSerializer
from api.serializers.profile import ProfileBasicSerializer
from api.models.event import TYPE_CHOICES
from api.models.project import STATE_CHOICES

SEARCH_TYPES = ["all", "project", "profile", "event"]
SEARCH_TYPE_CHOICES = [(SEARCH_TYPES[i], str(i))
                       for i in range(len(SEARCH_TYPES))]


class SearchRequestSerializer(serializers.Serializer):
    keyword = serializers.CharField(min_length=2)
    search_type = serializers.ChoiceField(
        choices=SEARCH_TYPE_CHOICES, allow_null=True, required=False,
        default="all")

    profile_affiliations = serializers.CharField(allow_null=True,
                                                 required=False)
    profile_expertise = serializers.CharField(allow_null=True, required=False)

    event_date_after = serializers.DateField(allow_null=True, required=False)
    event_date_before = serializers.DateField(allow_null=True, required=False)
    event_deadline_after = serializers.DateField(
        allow_null=True, required=False)
    event_deadline_before = serializers.DateField(
        allow_null=True, required=False)
    event_type = serializers.ChoiceField(
        choices=TYPE_CHOICES, allow_null=True, required=False)

    project_due_date_after = serializers.DateField(
        allow_null=True, required=False)
    project_due_date_before = serializers.DateField(
        allow_null=True, required=False)
    project_event = serializers.IntegerField(allow_null=True, required=False)
    project_event_title = serializers.CharField(allow_null=True,
                                                required=False)
    project_state = serializers.ChoiceField(
        choices=STATE_CHOICES, allow_null=True, required=False)
    project_tags = serializers.ListField(
        child=serializers.IntegerField(), allow_null=True, required=False)


class SearchResponseSerializer(serializers.Serializer):
    events = EventSerializer(read_only=True, many=True)
    projects = ProjectGETPublicSerializer(read_only=True, many=True)
    profiles = ProfileBasicSerializer(read_only=True, many=True)

    class Meta:
        fields = ['events', 'projects', 'profiles']



from django.contrib.auth.models import User

from api.models.event import Event
from rest_framework import viewsets
from rest_framework import permissions
from api.serializers.event import EventSerializer
from django_filters.rest_framework import DjangoFilterBackend
from api.views.feed import add_activity_to_feed


class EventViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = Event.objects.all()
    serializer_class = EventSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['event_type']

    def perform_create(self, serializer):
        serializer.save()

        '''
            Adds the events to admin feed.
        '''

        admin = User.objects.get(id=1)
        id = serializer.data['id']
        activity_data = {'actor': admin.username,
                         'verb': serializer.data['title'] + ' is created.',
                         'object': id,
                         'type': 'event',
                         'foreign_id': 'event:' + str(id),
                         'event': serializer.data,
                         }
        add_activity_to_feed('admin', activity_data)



from rest_framework import serializers
from api.models.tag import Tag


class TagSerializer(serializers.ModelSerializer):
    """
    Tag serializer
    """

    class Meta:
        model = Tag
        fields = '__all__'



from django.db import models

REPORT_TYPES = ["Disturbing other users",
                "Sharing unrelated or disturbing posts",
                "Spam", "Fake Profile", "Stolen Account"]
TYPE_CHOICES = [(REPORT_TYPES[i], str(i)) for i in range(len(REPORT_TYPES))]


class Report(models.Model):
    """
    Report model for project files.
    """
    reported_user = models.ForeignKey(
        'auth.User', related_name='reports', on_delete=models.CASCADE)
    report_type = models.CharField(choices=TYPE_CHOICES,
                                   default="Spam", max_length=100)
    description = models.CharField(max_length=1000, null=True)
    timestamp = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ['reported_user']



from rest_framework import serializers
from api.models.following import Following, FollowRequest
from api.serializers.user import UserPrivateSerializer, \
    UserNotificationSerializer


class FollowingSerializer(serializers.HyperlinkedModelSerializer):
    """
    Following serializer
    """
    from_user = serializers.IntegerField(source='from_user.id')

    class Meta:
        model = Following
        fields = ("id", "from_user", "created")

    def to_native(self, value):
        return self.parent.to_native(value)


class FollowerSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follower serializer
    """
    to_user = serializers.IntegerField(source='to_user.id')

    class Meta:
        model = Following
        fields = ("id", "to_user", "created")

    def to_native(self, value):
        return self.parent.to_native(value)


class FollowSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follow serializer
    """
    from_user = UserPrivateSerializer(read_only=True)
    to_user = UserPrivateSerializer(read_only=True)

    class Meta:
        model = Following
        fields = ("id", "from_user", "to_user", "created")


class FollowBasicSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follow serializer
    """
    from_user = UserNotificationSerializer(read_only=True)
    to_user = UserNotificationSerializer(read_only=True)

    class Meta:
        model = Following
        fields = ("id", "from_user", "to_user", "created")


class FollowPostSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follow serializer
    """
    to_user = serializers.IntegerField(source='to_user.id')

    class Meta:
        model = Following
        fields = ("to_user", "created")


class FollowRequestSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follow request serializer.
    """
    req_from_user = UserPrivateSerializer(read_only=True)
    req_to_user = UserPrivateSerializer(read_only=True)

    class Meta:
        model = FollowRequest
        fields = ("id", "req_from_user", "req_to_user", "created")


class FollowRequestPostSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follow request serializer.
    """
    req_to_user = serializers.IntegerField(source='req_to_user.id')

    class Meta:
        model = FollowRequest
        fields = ("req_to_user", "created")


class IncomingFollowRequestSerializer(serializers.HyperlinkedModelSerializer):
    """
    Follow request serializer.
    """
    from_user = serializers.IntegerField(source='req_from_user.id')

    class Meta:
        model = FollowRequest
        fields = ("id", "from_user", "created")

    def to_native(self, value):
        return self.parent.to_native(value)


class FollowRequestNotificationSerializer(serializers.
                                          HyperlinkedModelSerializer):
    """
    Follow request serializer.
    """
    req_from_user = UserNotificationSerializer(read_only=True)
    req_to_user = UserNotificationSerializer(read_only=True)

    class Meta:
        model = FollowRequest
        fields = ("id", "req_from_user", "req_to_user", "created")



# Generated by Django 3.1.4 on 2021-01-17 20:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0020_merge_20210117_1732'),
        ('api', '0020_auto_20210117_1512'),
    ]

    operations = [
    ]



from api.models.event import Event
from rest_framework import serializers


class EventSerializer(serializers.ModelSerializer):
    """
    Event serializer
    """
    class Meta:
        model = Event
        fields = '__all__'



# Generated by Django 3.1.3 on 2020-12-12 13:33

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0008_auto_20201206_2028'),
    ]

    operations = [
        migrations.CreateModel(
            name='FollowRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('req_from_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sent_requests', to=settings.AUTH_USER_MODEL)),
                ('req_to_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='incoming_requests', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.CreateModel(
            name='Following',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('from_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='following', to=settings.AUTH_USER_MODEL)),
                ('to_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='followers', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.AddConstraint(
            model_name='followrequest',
            constraint=models.UniqueConstraint(fields=('req_from_user', 'req_to_user'), name='unique_request'),
        ),
        migrations.AddConstraint(
            model_name='following',
            constraint=models.UniqueConstraint(fields=('from_user', 'to_user'), name='unique_followers'),
        ),
    ]



from rest_framework import viewsets
from rest_framework import permissions
from api.models.publication import Publication
from api.serializers.publication import PublicationSerializer
from scholarly import scholarly
from django.shortcuts import get_object_or_404
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework import status
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from django_filters.rest_framework import DjangoFilterBackend
from django.contrib.auth.models import User

author_param = openapi.Parameter(
    'author_id', openapi.IN_QUERY,
    description="Google Scholar id of author",
    type=openapi.TYPE_STRING)


class PublicationViewSet(viewsets.ModelViewSet):

    queryset = Publication.objects.all()
    serializer_class = PublicationSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['owner__id']

    @swagger_auto_schema(
        method='post', manual_parameters=[author_param]
    )
    @action(detail=False, methods=['post'], name='add_publications',
            serializer_class=None)
    def add_publications(self, request):
        owner_id = self.request.user.id
        owner = get_object_or_404(User, pk=owner_id)
        Publication.objects.filter(
            owner=owner
        ).delete()
        author_id = request.GET.get('author_id', None)
        author_basic = scholarly.search_author_id(author_id)
        author = scholarly.fill(author_basic)
        data = {}
        for publication in author['publications'][:25]:
            publication_info = publication['bib']
            if 'title' in publication_info:
                data['title'] = publication_info['title']
            if 'pub_year' in publication_info:
                data['publication_year'] = publication_info['pub_year']
            if 'author_pub_id' in publication:
                elem = publication['author_pub_id'].split(':')
                base_link = "https://scholar.google.com/citations?user={}" + \
                    "#d=gs_md_cita-d&u=%2Fcitations%3Fview_op%3D" +\
                    "view_citation%26user%3D{}%26citation_for_view%3D{}%3A{}"
                data['link'] = base_link.format(
                    elem[0], elem[0], elem[0], elem[1])
            if 'num_citations' in publication:
                data['citation_number'] = publication['num_citations']
            data['owner'] = owner_id

            serializer = PublicationSerializer(data=data)
            serializer.is_valid(raise_exception=True)
            Publication.objects.create(
                **serializer.validated_data)

        return Response(status=status.HTTP_201_CREATED)



from django.db import models
from .project import Project
from django.utils import timezone
from api.models.collaboration_invite import CollaborationInvite


class CollaborationRequest(models.Model):
    from_user = models.ForeignKey(
        'auth.User',
        on_delete=models.CASCADE,
        related_name="collab_request_sent",
    )
    to_user = models.ForeignKey(
        'auth.User',
        on_delete=models.CASCADE,
        related_name="collab_request_received",
    )
    to_project = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name="collab_request_of_project"
    )
    message = models.CharField(max_length=200, blank=True, default='')
    created = models.DateTimeField(blank=True, null=True)
    rejected = models.DateTimeField(blank=True, null=True)

    class Meta:
        ordering = ['created']

    def accept(self):
        self.to_project.members.add(self.from_user)
        self.delete()
        CollaborationInvite.objects.filter(
            from_user=self.to_user,
            to_user=self.from_user,
            to_project=self.to_project
        ).delete()
        # TODO add notification signal

    def reject(self):
        self.rejected = timezone.now()
        self.save()
        # TODO add notification signal



# Generated by Django 3.1.4 on 2021-01-17 13:57

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0015_publication'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publication',
            name='profile',
        ),
        migrations.AddField(
            model_name='publication',
            name='owner',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.CASCADE, related_name='publication_owner', to='auth.user'),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.1.4 on 2021-01-17 12:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0019_auto_20210117_1056'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tag',
            name='name',
            field=models.CharField(max_length=500, unique=True),
        ),
    ]



from rest_framework import viewsets
from rest_framework import permissions
from api.models.profile import Profile
from api.permission import IsOwnerOrReadOnly, ProfileDeletion
from api.serializers.profile import ProfileFullSerializer
from api.serializers.profile import ProfileBasicSerializer
from api.serializers.profile import ProfilePrivateSerializer
from api.serializers.profile import ProfilePictureSerializer
from django_filters.rest_framework import DjangoFilterBackend
from django.http import FileResponse
from rest_framework.response import Response
from rest_framework import status
from rest_framework.parsers import MultiPartParser, FormParser
from api.utils import get_is_following


class ProfileViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = Profile.objects.all()
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          IsOwnerOrReadOnly, ProfileDeletion]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['owner__id']

    def retrieve(self, request, *args, **kwargs):
        self.accessed_profile = self.get_object()
        serializer = self.get_serializer(self.accessed_profile)
        return Response(serializer.data)

    def get_serializer_class(self):
        this_user = self.request.user
        is_get = self.action == 'retrieve'
        is_list = self.action == 'list'
        if this_user.is_staff:
            return ProfileFullSerializer
        elif is_get:
            is_public = self.accessed_profile.is_public
            accessed_user = self.accessed_profile.owner
            is_following = get_is_following(this_user, accessed_user)
            if self.accessed_profile.owner == this_user:
                return ProfileFullSerializer
            elif is_public or is_following:
                return ProfileBasicSerializer
            else:
                return ProfilePrivateSerializer
        elif is_list:
            return ProfilePrivateSerializer
        else:
            return ProfileFullSerializer

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)


class ProfilePictureViewSet(viewsets.GenericViewSet):
    queryset = Profile.objects.all()
    serializer_class = ProfilePictureSerializer
    parser_classes = (MultiPartParser, FormParser)
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          IsOwnerOrReadOnly]

    def retrieve(self, request, pk=None):
        picture = self.get_object().profile_picture
        if picture:
            return FileResponse(picture)
        else:
            return Response("Profile picture is not found.",
                            status.HTTP_404_NOT_FOUND)

    def update(self, request, pk=None):
        data = {"profile_picture": request.data["profile_picture"]}
        profile = self.get_object()
        serializer = ProfilePictureSerializer(profile, data=data, partial=True)
        if profile.profile_picture:
            profile.profile_picture.delete()
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def destroy(self, request, pk=None):
        if self.get_object().profile_picture:
            self.get_object().profile_picture.delete()
            return Response("Profile picture is removed.", status.HTTP_200_OK)
        else:
            return Response("Profile picture is not found.",
                            status.HTTP_404_NOT_FOUND)



from api.models.collaboration_request import CollaborationRequest
from rest_framework import serializers


class CollaborationRequestSerializer(serializers.ModelSerializer):
    """
    Collaboration serializer
    """
    to_user = serializers.ReadOnlyField(source='to_user.id')
    from_user = serializers.ReadOnlyField(source='from_user.id')

    class Meta:
        model = CollaborationRequest
        fields = '__all__'



from api.models.comment import Comment
from api.models.profile import Profile
from django.contrib.auth.models import User
from rest_framework import viewsets
from rest_framework import permissions
from api.permission import CommentPermission
from api.serializers.comment import CommentSerializer, \
    CommentUpdateSerializer, CommentFeedSerializer, \
    CommentCreateSerializer
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.response import Response
from rest_framework import status
from api.utils import get_is_following, get_is_collaborator
from notifications.signals import notify
from django.db.models import Q

from api.views.feed import add_activity_to_feed


class CommentViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = Comment.objects.all()
    serializer_class = CommentSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          CommentPermission]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['to_user', 'from_user']

    def get_serializer_class(self):
        if self.action in ['update', 'partial_update']:
            return CommentUpdateSerializer
        elif self.action in ['create']:
            return CommentCreateSerializer
        else:
            return CommentSerializer

    def perform_create(self, serializer):
        serializer.save(from_user=self.request.user)

    def create(self, request, *args, **kwargs):
        '''
        Users can only comment on other user that they
        colllaborate on a project.
        '''
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.validated_data['from_user'] = self.request.user
        to_user = User.objects.get(
            id=serializer.validated_data['to_user'].id)
        serializer.validated_data['to_user'] = to_user

        is_collaborator = get_is_collaborator(self.request.user, to_user)

        if is_collaborator and \
                self.request.user.id != to_user.id:
            comment = Comment.objects.create(**serializer.validated_data)

            """
                Comment Notification
            """

            notify.send(sender=self.request.user,
                        verb="commented about you",
                        recipient=to_user,
                        target=comment,
                        description="Comment"
                        )

            """
                Adds the comment to user feed.
            """

            comment_serializer = CommentFeedSerializer(comment).data
            activity_data = {'actor': str(self.request.user),
                             'verb': 'commented {}'.format(
                                 comment_serializer['to_user']['username']),
                             'type': 'comment',
                             'object': comment_serializer['id'],
                             'foreign_id': 'comment:' +
                                           str(comment_serializer['id']),
                             'comment': comment_serializer,
                             }
            add_activity_to_feed(self.request.user, activity_data)

            return Response(serializer.data, status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def retrieve(self, request, *args, **kwargs):
        '''
        Users can see comments of a user that
        has a public profile or that they follow.
        '''
        self.accessed_comment = self.get_object()
        serializer = self.get_serializer(self.accessed_comment)

        to_user = self.accessed_comment.to_user
        is_public = Profile.objects.get(owner=to_user).is_public
        is_following = get_is_following(self.request.user, to_user)
        if is_public or is_following or \
                self.request.user.id == to_user.id:
            return Response(serializer.data)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    def list(self, request, *args, **kwargs):
        '''
        Users can list comments that belong to
        users with public profiles,
        themselves,
        someone they follow
        or comments they created.
        '''
        user = request.user
        queryset = self.filter_queryset(self.get_queryset())
        if user.is_anonymous:
            queryset = queryset.filter(to_user__profile__is_public__exact=True)
        elif not user.is_staff:
            queryset = queryset. \
                       filter(Q(to_user__profile__is_public__exact=True) |
                              Q(to_user__exact=user.id) |
                              Q(to_user__followers__exact=user.id) |
                              Q(from_user__exact=user.id))

        queryset = queryset.distinct()
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)



from django.db import models
import datetime
from .event import Event
from .tag import Tag

STATES = ["draft", "inviting collaborators", "open for collaborators",
          "in progress", "submitted to event", "published", "cancelled",
          "done", "reopened"]
STATE_CHOICES = [(STATES[i], str(i)) for i in range(len(STATES))]
PROJECT_TYPES = ["conference", "instutution", "journal"]
TYPE_CHOICES = [(PROJECT_TYPES[i], str(i)) for i in range(len(PROJECT_TYPES))]


class Project(models.Model):
    """
    Project model for projects.
    """
    created = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=500, blank=False)
    description = models.TextField(default='')
    requirements = models.TextField(default='')
    is_public = models.BooleanField(default=False)
    owner = models.ForeignKey(
        'auth.User', related_name='owner', on_delete=models.CASCADE)
    members = models.ManyToManyField('auth.User', related_name='members',
                                     blank=True)
    state = models.CharField(choices=STATE_CHOICES,
                             default="draft",
                             max_length=100)
    project_type = models.CharField(choices=TYPE_CHOICES,
                                    default="conference", max_length=100)
    due_date = models.DateField(default=datetime.date.today)
    event = models.ForeignKey(
        Event, blank=True, on_delete=models.SET_NULL, null=True)
    tags = models.ManyToManyField(Tag, blank=True)

    class Meta:
        ordering = ['created']



from django.db.models import Q
from api.models.profile import Profile
from api.models.project import Project
from api.models.event import Event
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework import generics
from api.serializers.search import SearchRequestSerializer
from api.serializers.event import EventSerializer
from api.serializers.project import ProjectPrivateSerializer
from api.serializers.profile import ProfilePrivateSerializer
from datamuse import datamuse
from api.models.following import Following
from django.contrib.auth.models import AnonymousUser
import re
from api.utils import get_user_rating
import math


class SearchGenericAPIView(generics.GenericAPIView):
    """
        Execute a search with a string
    """
    serializer_class = SearchRequestSerializer

    def post(self, request: Request, *args, **kwargs):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)

        req_data = serializer.validated_data

        keyword_req = req_data['keyword']
        search_type = req_data['search_type']

        api = datamuse.Datamuse()
        keyword_list = api.words(rel_jja=keyword_req, max=5)
        keywords = [keyword_req]
        for keyword_dict in keyword_list:
            keywords.append(keyword_dict['word'])

        res = {}

        query_events = []
        query_profiles = []
        query_followed_profiles = []
        query_private_profiles = []
        query_projects = []
        query_private_projects = []

        isGuest = isinstance(request.user, AnonymousUser)
        search_projects = search_type in ["all", "project"]
        search_profiles = search_type in ["all", "profile"]
        search_events = search_type in ["all", "event"]

        public_project_filters = set(["project_due_date_after",
                                      "project_due_date_before",
                                      "project_event", "project_event_title"])
        any_public_project_filter = len(
            public_project_filters.intersection(set(req_data.keys()))) > 0

        public_profile_filters = set(["profile_affiliations",
                                      "profile_expertise"])
        any_public_profile_filter = len(
            public_profile_filters.intersection(set(req_data.keys()))) > 0

        query_following = []
        if not isGuest:
            query_following = list(map(lambda following: following.to_user,
                                       Following.objects.
                                       filter(from_user=request.user)))

        for keyword in keywords:
            if search_events:
                q = Q(title__icontains=keyword) | Q(
                    description__icontains=keyword)

                if "event_date_after" in req_data:
                    q &= Q(date__gte=req_data["event_date_after"])
                if "event_date_before" in req_data:
                    q &= Q(date__lte=req_data["event_date_before"])
                if "event_deadline_after" in req_data:
                    q &= Q(deadline__gte=req_data["event_deadline_after"])
                if "event_deadline_before" in req_data:
                    q &= Q(deadline__lte=req_data["event_deadline_before"])
                if "event_type" in req_data:
                    q &= Q(event_type=req_data["event_type"])

                query_events += Event.objects.filter(q)

            if search_projects:
                q = Q()

                if isGuest:
                    q = Q(is_public=True)
                else:
                    q = (Q(is_public=True) | Q(
                        members__id=request.user.id))

                if "project_due_date_after" in req_data:
                    q &= Q(due_date__gte=req_data["project_due_date_after"])
                if "project_due_date_before" in req_data:
                    q &= Q(due_date__lte=req_data["project_due_date_before"])
                if "project_event" in req_data:
                    q &= Q(event__id=req_data["project_event"])
                if "project_event_title" in req_data:
                    q &= Q(
                        event__title__icontains=req_data["project_event_title"]
                        )
                if "project_state" in req_data:
                    q &= Q(state=req_data["project_state"])
                if "project_tags" in req_data:
                    print(req_data["project_tags"])
                    q &= Q(tags__id__in=req_data["project_tags"])

                q &= (Q(name__icontains=keyword) |
                      Q(description__icontains=keyword) |
                      Q(project_type__icontains=keyword) |
                      Q(event__title__icontains=keyword) |
                      Q(event__description__icontains=keyword))

                query_projects += Project.objects.filter(q)

                if not any_public_project_filter:
                    q = Q()
                    if isGuest:
                        q = Q(is_public=False)
                    else:
                        q = (Q(is_public=False) & ~Q(
                            members__id=request.user.id))

                    if "project_state" in req_data:
                        q &= Q(state=req_data["project_state"])
                    if "project_tags" in req_data:
                        q &= Q(tags__id__in=req_data["project_tags"])

                    q &= (Q(name__icontains=keyword) |
                          Q(description__icontains=keyword))

                    query_private_projects += Project.objects.filter(q)

            if search_profiles:
                q = Q(is_public=True)
                q &= (Q(name__icontains=keyword) |
                      Q(middle_name__icontains=keyword) |
                      Q(last_name__icontains=keyword) |
                      Q(expertise__icontains=keyword) |
                      Q(interests__icontains=keyword) |
                      (Q(share_bio=True) & Q(bio__icontains=keyword)) |
                      (Q(share_affiliations=True) &
                       Q(affiliations__icontains=keyword)))

                if "profile_affiliations" in req_data:
                    q &= Q(
                        affiliations__icontains=req_data[
                            "profile_affiliations"])
                if "profile_expertise" in req_data:
                    q &= Q(expertise__icontains=req_data["profile_expertise"])

                query_profiles += Profile.objects.filter(q)

                if not isGuest:
                    q = (Q(is_public=False) & Q(owner__in=query_following))

                    q &= (Q(name__icontains=keyword) |
                          Q(middle_name__icontains=keyword) |
                          Q(last_name__icontains=keyword) |
                          Q(expertise__icontains=keyword) |
                          Q(interests__icontains=keyword) |
                          (Q(share_bio=True) & Q(bio__icontains=keyword)) |
                          (Q(share_affiliations=True) &
                           Q(affiliations__icontains=keyword)))

                    if "profile_affiliations" in req_data:
                        q &= Q(
                            affiliations__icontains=req_data[
                                "profile_affiliations"])
                    if "profile_expertise" in req_data:
                        q &= Q(
                            expertise__icontains=req_data["profile_expertise"])

                    query_followed_profiles += Profile.objects.filter(q)

                if not any_public_profile_filter:
                    q = Q()
                    if isGuest:
                        q = Q(is_public=False)
                    else:
                        q = (Q(is_public=False) & ~Q(
                            owner__in=query_following))
                    q &= (Q(name__icontains=keyword) |
                          Q(middle_name__icontains=keyword) |
                          Q(last_name__icontains=keyword))

                    query_private_profiles += Profile.objects.filter(q)

        # Fill the Response
        if search_events:
            query_events = list(set(query_events))

            event_serializer = EventSerializer(query_events, many=True,
                                               context={'request': request})

            res['events'] = event_serializer.data

        if search_projects:
            query_projects = list(set(query_projects))
            query_private_projects = list(set(query_private_projects))

            exps = []
            projects_all = query_projects + query_private_projects

            if not isGuest:
                profile = Profile.objects.get(owner_id=self.request.user.id)
                if profile.expertise:
                    exps = re.split('; |, |\n', profile.expertise)
                    exps = [r.strip() for r in exps]

                project_score = {}

                for project in projects_all:
                    rating = get_user_rating(project.owner_id)
                    scaled_rating = rating/10.0 if rating else 0.5
                    project_score[project.id] = [1 + scaled_rating, 1, 1]

                    members = project.members.all()
                    members = [member.id for member in members]
                    if self.request.user.id in members:
                        project_score[project.id][2] = 5
                    else:
                        for followed in query_following:
                            if followed.id in members:
                                project_score[project.id][2] = 1.5
                                break
                    for x in exps:
                        if x in project.requirements:
                            project_score[project.id][1] += 1

                for project_id, scores in project_score.items():
                    project_score[project_id] = math.prod(scores)

                top_ids = [i for i in sorted(project_score,
                                             key=project_score.get,
                                             reverse=True)]

                projects_all = [project for pid in top_ids
                                for project in projects_all
                                if project.id == pid]

            private_project_serializer = ProjectPrivateSerializer(
                projects_all, many=True,
                context={'request': request})

            res['projects'] = private_project_serializer.data

        if search_profiles:
            query_profiles = list(set(query_profiles))
            query_followed_profiles = list(set(query_followed_profiles))
            query_private_profiles = list(set(query_private_profiles))

            profiles_all = query_profiles + query_followed_profiles + \
                query_private_profiles

            if not isGuest:
                query_following_ids = [
                    user.id for user in query_following]
                query_following_following = list(
                    map(lambda following: following.to_user,
                        Following.objects.
                        filter(from_user__in=query_following)))
                query_following_following_ids = [
                    user.id for user in query_following_following]
                query_following_following_ids = list(
                    set(query_following_following_ids) -
                    set(query_following_ids))

                profile_score = {}

                for profile in profiles_all:
                    rating = get_user_rating(profile.owner_id)
                    scaled_rating = rating/10.0 if rating else 0.5
                    profile_score[profile.id] = [1 + scaled_rating, 1]

                    if profile.owner_id in query_following_ids:
                        profile_score[profile.id][1] = 5
                    elif profile.owner_id in query_following_following_ids:
                        profile_score[profile.id][1] = 3

                for profile_id, scores in profile_score.items():
                    profile_score[profile_id] = math.prod(scores)

                top_ids = [i for i in sorted(profile_score,
                                             key=profile_score.get,
                                             reverse=True)]

                profiles_all = [profile for pid in top_ids
                                for profile in profiles_all
                                if profile.id == pid]

            private_profile_serializer = ProfilePrivateSerializer(
                profiles_all, many=True,
                context={'request': request})

            res['profiles'] = private_profile_serializer.data

        return Response(data=res)



from django.db import models
from django.core.validators import MaxValueValidator, MinValueValidator


class Rating(models.Model):
    """
    Rating model for ratings.
    """
    from_user = models.ForeignKey('auth.User',
                                  related_name="ratings_made",
                                  on_delete=models.CASCADE,
                                  )
    to_user = models.ForeignKey('auth.User',
                                related_name="ratings_received",
                                on_delete=models.CASCADE,
                                )
    created = models.DateTimeField(auto_now_add=True)
    rating = models.IntegerField(validators=[MaxValueValidator(10),
                                             MinValueValidator(0)])

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['from_user',
                                            'to_user'],
                                    name="unique_rating")
        ]
        ordering = ['created']



# Generated by Django 3.1.4 on 2020-12-17 11:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0013_merge_20201215_2314'),
    ]

    operations = [
        migrations.AlterField(
            model_name='followrequest',
            name='req_to_user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='follow_requests', to=settings.AUTH_USER_MODEL),
        ),
    ]



from rest_framework import serializers
from api.models.milestone import Milestone


class MilestoneSerializer(serializers.ModelSerializer):
    """
    Milestone serializer
    """

    class Meta:
        model = Milestone
        fields = '__all__'



# Generated by Django 3.1.3 on 2020-12-06 14:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0002_collaborationrequest'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collaborationrequest',
            name='to_project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='collab_request_of_project', to='api.project'),
        ),
    ]



# Generated by Django 3.1.3 on 2020-12-04 14:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='project',
            name='events',
        ),
        migrations.RemoveField(
            model_name='tag',
            name='projects',
        ),
        migrations.AddField(
            model_name='project',
            name='event',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='api.event'),
        ),
        migrations.AddField(
            model_name='project',
            name='tags',
            field=models.ManyToManyField(blank=True, to='api.Tag'),
        ),
    ]



from django.db import models

TYPES = ["journal submission", "academic conference", "funded project"]
TYPE_CHOICES = [(TYPES[i], str(i)) for i in range(len(TYPES))]


class Event(models.Model):
    """
    Event model
    """
    title = models.CharField(max_length=100, blank=False)
    description = models.CharField(max_length=200, blank=True, default='')
    deadline = models.DateField(blank=False)
    date = models.DateField(blank=False)
    event_type = models.CharField(
        choices=TYPE_CHOICES, blank=False, max_length=50)
    url = models.URLField(blank=True, default='')

    class Meta:
        ordering = ['date']



# Generated by Django 3.1.3 on 2020-12-06 18:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0002_auto_20201204_1708'),
        ('api', '0003_auto_20201206_1712'),
    ]

    operations = [
    ]



# Generated by Django 3.1.4 on 2021-01-21 13:41

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('api', '0021_merge_20210117_2302'),
    ]

    operations = [
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('report_type', models.CharField(choices=[('Disturbing other users', '0'), ('Sharing unrelated or disturbing posts', '1'), ('Spam', '2'), ('Fake Profile', '3'), ('Stolen Account', '4')], default='Spam', max_length=100)),
                ('description', models.CharField(max_length=1000, null=True)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('reported_user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reports', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['reported_user'],
            },
        ),
    ]



from django.db import models


class Comment(models.Model):
    """
    Comment model for comments.
    """
    from_user = models.ForeignKey('auth.User',
                                  related_name="comments_made",
                                  on_delete=models.CASCADE,
                                  )
    to_user = models.ForeignKey('auth.User',
                                related_name="comments_received",
                                on_delete=models.CASCADE,
                                )
    created = models.DateTimeField(auto_now_add=True)
    comment = models.CharField(max_length=1000)

    class Meta:
        ordering = ['created']



from rest_framework import viewsets
from django.contrib.auth.models import User
from api.models.profile import Profile
from api.serializers.user import UserFullSerializer
from api.serializers.user import UserBasicSerializer
from api.serializers.user import UserPrivateSerializer
from rest_framework.response import Response
from api.utils import get_is_following, get_user_rating
from rest_framework.decorators import action
from django.shortcuts import get_object_or_404
from api.models.project import Project
from api.models.following import Following
from api.models.collaboration_invite import CollaborationInvite
import math
import re
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from django.db.models import Case, When
from django_filters.rest_framework import DjangoFilterBackend

project_param = openapi.Parameter(
    'project_id', openapi.IN_QUERY,
    description="Project id to get recommendations",
    type=openapi.TYPE_INTEGER)
count_param = openapi.Parameter(
    'project_count', openapi.IN_QUERY,
    description="Number of users requested",
    type=openapi.TYPE_INTEGER)


class UserViewSet(viewsets.ReadOnlyModelViewSet):
    """
    This viewset automatically provides `list` and `detail` actions.
    """
    queryset = User.objects.all()
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['username']

    def retrieve(self, request, *args, **kwargs):
        self.accessed_user = self.get_object()
        serializer = self.get_serializer(self.accessed_user)
        return Response(serializer.data)

    def get_serializer_class(self):
        this_user = self.request.user
        is_get = self.action == 'retrieve'
        is_list = self.action == 'list'
        if this_user.is_staff:
            return UserFullSerializer
        elif is_get:
            is_public = Profile.objects.get(owner=self.accessed_user).is_public
            is_following = get_is_following(this_user, self.accessed_user)
            if self.accessed_user == this_user:
                return UserFullSerializer
            elif is_public or is_following:
                return UserBasicSerializer
            else:
                return UserPrivateSerializer
        elif is_list:
            return UserPrivateSerializer
        else:
            return UserFullSerializer

    @swagger_auto_schema(
        method='get', manual_parameters=[project_param, count_param]
    )
    @action(detail=False, methods=['GET'],
            name='get_collaborator_recommendation',
            serializer_class=None)
    def get_collaborator_recommendation(self, request):
        project_id = request.GET.get('project_id', None)
        project_count = request.GET.get('project_count', None)
        project = get_object_or_404(Project, pk=project_id)
        project_owner = project.owner_id
        reqs = []
        if project.requirements:
            reqs = re.split('; |, |\n', project.requirements)
            reqs = [r.strip() for r in reqs]

        """
            value of the dictionary items will be a list of length 3:
            [rating_score, expertise_score, following_score]

            The end score calculation will be the following:
            rating_score*expertise_score*following_score

            Base value for rating_score is 1, rating/10 is added to this value.
            If someone has no rating it is treated as a rating of 5.

            Default score for expertise is 1 and is incremented by 1
            with each match with requirements.

            Following_score is 1.2 for the users that the owner follows,
            1 otherwise.
        """

        user_score = {}

        profiles = Profile.objects.all()
        for profile in profiles:
            rating = get_user_rating(profile.owner_id)
            scaled_rating = rating/10.0 if rating else 0.5
            user_score[profile.owner_id] = [1 + scaled_rating, 1, 1]

        for keyword in reqs:
            profiles = Profile.objects.filter(
                expertise__icontains=keyword)
            for profile in profiles:
                user_score[profile.owner_id][1] += 1

        followed_users = Following.objects.filter(from_user=project_owner)
        for followed in followed_users:
            user_score[followed.to_user.id][2] = 1.2

        for user_id, scores in user_score.items():
            user_score[user_id] = math.prod(scores)

        # pop the project members out of the list
        for member in project.members.all():
            user_score.pop(member.id, None)

        invited_users = CollaborationInvite.objects.filter(
            from_user=project_owner,
            to_project=project)
        # pop the invited users out of the list
        for invite in invited_users:
            user_score.pop(invite.to_user.id, None)

        top_ids = []
        if project_count:
            top_ids = [i for i in sorted(user_score,
                                         key=user_score.get,
                                         reverse=True)[:int(project_count)]]
        else:
            top_ids = [i for i in sorted(user_score,
                                         key=user_score.get,
                                         reverse=True)]

        preserved = Case(*[When(id=pk, then=pos)
                           for pos, pk in enumerate(top_ids)])
        users = User.objects.filter(id__in=top_ids).order_by(preserved)
        serializer_context = {'request': request}
        serializer = UserBasicSerializer(
            users, many=True, context=serializer_context)
        return Response(serializer.data)



# Generated by Django 3.1.4 on 2020-12-20 20:13

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0014_auto_20201217_1125'),
    ]

    operations = [
        migrations.CreateModel(
            name='Publication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=1000)),
                ('publication_year', models.IntegerField(default=0)),
                ('abstract', models.TextField(default='')),
                ('authors', models.TextField(default='')),
                ('journal', models.TextField(default='')),
                ('citation_number', models.IntegerField(default=0)),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.profile')),
            ],
            options={
                'ordering': ['citation_number'],
            },
        ),
    ]



from django.db import models
import random


def random_color():
    return random.randint(0, 9)


class Tag(models.Model):
    """
    Tag model for project tags.
    """
    name = models.CharField(max_length=500, blank=False, unique=True)
    color = models.IntegerField(default=random_color)

    class Meta:
        ordering = ['name']



import sendgrid
import os

sg = sendgrid.SendGridAPIClient(api_key=os.environ.get("SENDGRID_APIKEY"))


def send_mail(name, url):
    data = {
        "from": {
            "email": "bounswe2020group3@gmail.com",
            "name": "Paper Layer"
        },
        "personalizations": [
            {
                "to": [
                    {
                        "email": "f.cansever67@gmail.com"
                    }
                ],
                "dynamic_template_data": {
                    "name": name,
                    "notification_url": url
                }
            }
        ],
        "template_id": "d-8f1578c0872a4e2680862372bd0f5bf1"
    }

    sg.client.mail.send.post(request_body=data)



from api.models.collaboration_invite import CollaborationInvite
from rest_framework import serializers


class CollaborationInviteSerializer(serializers.ModelSerializer):
    """
    Collaboration serializer
    """
    to_user = serializers.ReadOnlyField(source='to_user.id')
    from_user = serializers.ReadOnlyField(source='from_user.id')

    class Meta:
        model = CollaborationInvite
        fields = '__all__'


class CollaborationInvitePOSTSerializer(serializers.ModelSerializer):
    """
    Post serializer, includes only user
    """
    from_user = serializers.ReadOnlyField(source='from_user.id')

    class Meta:
        model = CollaborationInvite
        fields = '__all__'



# Generated by Django 3.1.4 on 2021-01-24 10:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0021_merge_20210117_2302'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publication',
            name='abstract',
        ),
        migrations.RemoveField(
            model_name='publication',
            name='authors',
        ),
        migrations.RemoveField(
            model_name='publication',
            name='journal',
        ),
        migrations.AddField(
            model_name='publication',
            name='link',
            field=models.TextField(default=''),
        ),
    ]



from rest_framework import viewsets
from rest_framework import permissions
from api.permission import IsMemberOrReadOnly
from api.serializers.project import ProjectPublicSerializer
from api.serializers.project import ProjectGETPublicSerializer
from api.serializers.project import ProjectPrivateSerializer
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.response import Response
from notifications.signals import notify
from api.models.project import Project
from api.views.feed import add_activity_to_feed
from rest_framework.decorators import action
from api.models.profile import Profile
from api.models.following import Following
from api.models.collaboration_request import CollaborationRequest
import math
import re
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from django.db.models import Case, When
from api.utils import get_user_rating

user_param = openapi.Parameter(
    'user_id', openapi.IN_QUERY,
    description="User id to get recommendations",
    type=openapi.TYPE_INTEGER)
count_param = openapi.Parameter(
    'user_count', openapi.IN_QUERY,
    description="Number of projects requested",
    type=openapi.TYPE_INTEGER)


class ProjectViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = Project.objects.all()
    serializer_class = ProjectPublicSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly,
                          IsMemberOrReadOnly]
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['owner__id', 'members__id']

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user, members=[self.request.user])

    def retrieve(self, request, *args, **kwargs):
        self.accessed_project = self.get_object()
        serializer = self.get_serializer(self.accessed_project)
        return Response(serializer.data)

    def get_serializer_class(self):
        this_user = self.request.user
        is_get = self.action == 'retrieve'
        is_list = self.action == 'list'
        if this_user.is_staff:
            return ProjectGETPublicSerializer
        elif is_get and self.accessed_project.owner == this_user:
            return ProjectGETPublicSerializer
        elif is_get and this_user in self.accessed_project.members.all():
            return ProjectGETPublicSerializer
        elif is_get and self.accessed_project.is_public:
            return ProjectGETPublicSerializer
        elif is_get and not self.accessed_project.is_public:
            return ProjectPrivateSerializer
        elif is_list:
            return ProjectPrivateSerializer
        else:
            return ProjectPublicSerializer

    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)
        if self.action == 'create':
            project = Project.objects.get(id=response.data['id'])
            notify.send(sender=self.request.user,
                        verb="created a new Project {}".
                        format(response.data['name']),
                        recipient=self.request.user,
                        target=project,
                        description='Project')
            # send_mail(self.request.user)

            """
                Adds the project to the user feed
            """
            activity_data = {'actor': str(self.request.user),
                             'verb': 'created new Project called '
                                     + project.name,
                             'type': 'project',
                             'object': project.id,
                             'foreign_id': 'project:' + str(project.id),
                             'project': ProjectPublicSerializer(project).data,
                             }
            add_activity_to_feed(self.request.user, activity_data)

        if self.action == 'update':
            pass
        if self.action == 'destroy':
            pass
        return response

    @swagger_auto_schema(
        method='get', manual_parameters=[user_param, count_param]
    )
    @action(detail=False, methods=['GET'],
            name='get_project_recommendation',
            serializer_class=None)
    def get_project_recommendation(self, request):
        user_id = request.GET.get('user_id', None)
        user_count = request.GET.get('user_count', None)
        profile = Profile.objects.get(owner_id=user_id)
        exps = []
        if profile.expertise:
            exps = re.split('; |, |\n', profile.expertise)
            exps = [r.strip() for r in exps]

        """
            value of the dictionary items will be a list of length 3:
            [rating_score, requirement_score, following_score]

            The end score calculation will be the following:
            rating_score*requirement_score*following_score

            Base value for rating_score is 1, rating/10 is added to this value.
            If someone has no rating it is treated as a rating of 5.
            This is the rating of the owner of the project.

            Default score for requirement is 1 and is incremented by 1
            with each match with expertise.

            Following_score is 1.2 for the pprojects that contain members
            that the user follows, 1 otherwise.
        """

        project_score = {}

        followed_users = Following.objects.filter(from_user=user_id)
        projects = Project.objects.filter(
            state="open for collaborators").order_by('-created')
        for project in projects:
            rating = get_user_rating(project.owner_id)
            scaled_rating = rating/10.0 if rating else 0.5
            project_score[project.id] = [1 + scaled_rating, 1, 1]

            members = project.members.all()
            members = [member.id for member in members]
            for followed in followed_users:
                if followed.to_user.id in members:
                    project_score[project.id][2] = 1.2
                    break

        for keyword in exps:
            projects = Project.objects.filter(
                requirements__icontains=keyword,
                state="open for collaborators")
            for project in projects:
                project_score[project.id][1] += 1

        for project_id, scores in project_score.items():
            project_score[project_id] = math.prod(scores)

        member_projects = Project.objects.filter(
            members__id__in=[user_id])
        # pop the projects that the user is a member of out of the list
        for m_project in member_projects:
            project_score.pop(m_project.id, None)

        req_projects = CollaborationRequest.objects.filter(
            from_user=user_id)
        # pop the requested projects out of the list
        for req in req_projects:
            project_score.pop(req.to_project.id, None)

        top_ids = []
        if user_count:
            top_ids = [i for i in sorted(project_score,
                                         key=project_score.get,
                                         reverse=True)[:int(user_count)]]
        else:
            top_ids = [i for i in sorted(project_score,
                                         key=project_score.get,
                                         reverse=True)]

        preserved = Case(*[When(id=pk, then=pos)
                           for pos, pk in enumerate(top_ids)])
        projects = Project.objects.filter(id__in=top_ids).order_by(preserved)
        serializer_context = {'request': request}
        serializer = ProjectGETPublicSerializer(
            projects, many=True, context=serializer_context)
        return Response(serializer.data)



from notifications.base.models import AbstractNotification
from django.contrib.contenttypes.fields import GenericForeignKey


class Notification(AbstractNotification):
    user = GenericForeignKey('target_content_type', 'target_object_id')
    invite = GenericForeignKey('target_content_type', 'target_object_id')
    request = GenericForeignKey('target_content_type', 'target_object_id')
    follow = GenericForeignKey('target_content_type', 'target_object_id')
    project = GenericForeignKey('target_content_type', 'target_object_id')
    milestone = GenericForeignKey('target_content_type', 'target_object_id')
    following = GenericForeignKey('target_content_type', 'target_object_id')
    follow_request = GenericForeignKey('target_content_type',
                                       'target_object_id')
    comment = GenericForeignKey('target_content_type', 'target_object_id')
    rating = GenericForeignKey('target_content_type', 'target_object_id')

    class Meta(AbstractNotification.Meta):
        abstract = False
        app_label = 'api'



from django.db import models

GENDERS = ["male", "female", "do not want to share"]
GENDER_CHOICES = [(GENDERS[i], str(i)) for i in range(len(GENDERS))]


class Profile(models.Model):
    """
    Profile model for user profiles
    """
    created = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=100, blank=False, default='')
    middle_name = models.CharField(max_length=100, blank=True, default='')
    last_name = models.CharField(max_length=100, blank=False, default='')
    bio = models.CharField(max_length=1000, blank=True, default='')
    profile_picture = models.ImageField(null=True, blank=True, upload_to='pp/')
    birthday = models.DateField(blank=True, null=True)
    share_birthday = models.BooleanField(default=True)
    expertise = models.TextField(default='', blank=True)
    gender = models.CharField(choices=GENDER_CHOICES,
                              default="do not want to share", max_length=100)
    interests = models.TextField(default='', blank=True)
    affiliations = models.TextField(default='', blank=True)
    owner = models.ForeignKey(
        'auth.User', related_name='profile', on_delete=models.CASCADE)
    share_bio = models.BooleanField(default=True)
    share_gender = models.BooleanField(default=True)
    share_affiliations = models.BooleanField(default=True)
    is_public = models.BooleanField(default=False)

    class Meta:
        ordering = ['created']



from django_email_verification import urls as mail_urls
from django.contrib import admin
from .views.rating import RatingViewSet
from .views.comment import CommentViewSet
from django.contrib.auth import views as auth_views
from .views.search import SearchGenericAPIView
from .views.notification import NotificationViewSet
from .views.publication import PublicationViewSet
from django.urls import path
from django.conf.urls import include
from rest_framework.routers import DefaultRouter
from .views.feed import FeedViewSet
from .views.following import FollowingViewSet, FollowRequestViewSet
from .views.profile import ProfileViewSet, ProfilePictureViewSet
from .views.auth import RegisterGenericAPIView, LogoutGenericAPIView, AuthView
from .views.project import ProjectViewSet
from .views.milestone import MilestoneViewSet
from .views.tag import TagViewSet
from .views.user import UserViewSet
from .views.event import EventViewSet
from .views.file import FileViewSet
from .views.collaboration_request import CollaborationRequestViewSet
from .views.collaboration_invite import CollaborationInviteViewSet
from .views.report import ReportViewSet

router = DefaultRouter()
router.register(r'profiles', ProfileViewSet)
router.register(r'profile_picture', ProfilePictureViewSet)
router.register(r'users', UserViewSet)
router.register(r'projects', ProjectViewSet)
router.register(r'milestones', MilestoneViewSet)
router.register(r'tags', TagViewSet)
router.register(r'events', EventViewSet)
router.register(r'files', FileViewSet)
router.register(r'follow', FollowingViewSet)
router.register(r'follow_request', FollowRequestViewSet)
router.register(r'collaboration_requests', CollaborationRequestViewSet)
router.register(r'collaboration_invites', CollaborationInviteViewSet)
router.register(r'notifications', NotificationViewSet, basename='notification')
router.register(r'comments', CommentViewSet)
router.register(r'ratings', RatingViewSet)
router.register(r'feeds', FeedViewSet, basename='feed')
router.register(r'publications', PublicationViewSet)
router.register(r'reports', ReportViewSet)

urlpatterns = [
    path('', include(router.urls)),
    path('register/', RegisterGenericAPIView.as_view()),
    path('auth/', AuthView.as_view(), name='auth'),
    path('logout/', LogoutGenericAPIView.as_view()),
    path('reset_password/', auth_views.PasswordResetView.as_view(),
         name="reset_password"),
    path('reset_password_sent/', auth_views.PasswordResetDoneView.as_view(),
         name="password_reset_done"),
    path('reset/<uidb64>/<token>/',
         auth_views.PasswordResetConfirmView.as_view(),
         name="password_reset_confirm"),
    path('reset_password_complete/',
         auth_views.PasswordResetCompleteView.as_view(),
         name="password_reset_complete"),
    path('admin/', admin.site.urls),
    path('email/', include(mail_urls)),
    path('search/', SearchGenericAPIView.as_view(), name='search'),

]



# Generated by Django 3.1.3 on 2020-12-14 17:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0009_auto_20201209_1436'),
        ('api', '0009_auto_20201208_2042'),
    ]

    operations = [
    ]



from rest_framework import viewsets
from rest_framework import permissions
from api.models.tag import Tag
from api.serializers.tag import TagSerializer


class TagViewSet(viewsets.ModelViewSet):
    """
    This viewset automatically provides `list`, `create`, `retrieve`,
    `update` and `destroy` actions.
    """
    queryset = Tag.objects.all()
    serializer_class = TagSerializer
    permission_classes = [permissions.IsAuthenticatedOrReadOnly]



from django.db import models
from api.models.project import Project


class File(models.Model):
    """
    File model for project files.
    """
    file = models.FileField(blank=False, null=False,
                            upload_to='files/')
    remark = models.CharField(max_length=100)
    timestamp = models.DateTimeField(auto_now_add=True)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)

    class Meta:
        ordering = ['project']

    def delete(self, *args, **kwargs):
        if self.file:
            self.file.delete()
        super(File, self).delete(*args, **kwargs)



# Generated by Django 3.1.4 on 2020-12-28 12:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0016_auto_20201221_2047'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collaborationinvite',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='collaborationrequest',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='comment',
            name='comment',
            field=models.CharField(max_length=1000),
        ),
    ]



from setuptools import setup, find_packages

setup(
  name = 'django-paperlayer',         # How you named your package folder (MyLib)
  packages = ['django-paperlayer',
              'django-paperlayer/api',
              'django-paperlayer/api/migrations',
              'django-paperlayer/api/models',
              'django-paperlayer/api/serializers',
              'django-paperlayer/api/views',
              'django-paperlayer/backend',
              'django-paperlayer/templates',
              ],   # Chose the same as "name"
  version = '1.3',      # Start with a small number and increase it with every change you make
  license= 'MIT License',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'An simple academic platform',   # Give a short description about your library
  author = 'Paperlayer Dev Team',                   # Type in your name
  author_email = 'bounswe2020group3@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/bounswe/bounswe2020group3',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/bounswe/bounswe2020group3/archive/2.2.1.tar.gz',    # I explain this later on
  python_requires='>=3.6',
  install_requires=[            # I get to this in a second
      'Django>=3.1.3',
      'bandit>=1.6.2',
      'djangorestframework>=3.12.2',
      'drf-yasg>=1.20.0',
      'django-password-reset>=2.0',
      'gunicorn>=20.0.4',
      'django-filter>=2.4.0',
      'django-cors-headers>=3.5.0',
      'Pillow>=8.0.1',
      'python-datamuse>=1.3.0'
      'django-storages>=1.11.1',
      'boto3>=1.16.43',
      'django-email-verification>=0.0.7',
      'django-notifications-hq>=1.6.0',
      'sendgrid>=6.4.8',
      'stream-python>=5.0.0',
      'scholarly>=1.0.3'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.9',
  ],
)


from django.db import models
from api.models.project import Project


class Milestone(models.Model):
    """
    Milestone model for project milestones.
    """
    description = models.TextField(default='')
    date = models.DateField()
    project = models.ForeignKey(Project, on_delete=models.CASCADE,
                                related_name='milestones')

    class Meta:
        ordering = ['project']



# Generated by Django 3.1.5 on 2021-01-24 12:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0022_auto_20210124_1320'),
        ('api', '0022_report'),
    ]

    operations = [
    ]



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.authtoken.admin import User
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticatedOrReadOnly
from api.models.following import Following, FollowRequest
from api.permission import IsRequestSenderOrReceiver
from api.serializers.following import FollowSerializer, \
    FollowRequestSerializer, FollowPostSerializer, \
    FollowRequestPostSerializer, FollowBasicSerializer
from django.shortcuts import get_object_or_404
from rest_framework.response import Response
from rest_framework import status
from notifications.signals import notify

from api.serializers.user import UserNotificationSerializer
from api.views.feed import add_activity_to_feed, follow_user_timeline


class FollowingViewSet(viewsets.ModelViewSet):
    """
    ViewSet for follow.
    """

    permission_classes = (IsAuthenticatedOrReadOnly,)
    queryset = Following.objects.all()
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['from_user__id', 'to_user__id']

    def perform_create(self, serializer):
        serializer.save(from_user=self.request.user)

    def create(self, request, *args, **kwargs):
        serializer = FollowPostSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.validated_data['from_user'] = self.request.user
        to_user = User.objects.get(
            id=serializer.validated_data['to_user']['id'])
        serializer.validated_data['to_user'] = to_user
        follow = Following.objects.create(**serializer.validated_data)

        ''' Follow Notification '''
        ''' Target --> Follow '''
        notify.send(sender=self.request.user,
                    verb="followed you.",
                    recipient=to_user,
                    target=follow,
                    description="Follow"
                    )

        '''
            Adds the follow to the user feed
        '''
        follow_serializer = FollowBasicSerializer(follow).data
        activity_data = {'actor': str(self.request.user),
                         'verb': 'followed ' +
                                 follow_serializer['to_user']['username'],
                         'type': 'follow',
                         'object': follow_serializer['id'],
                         'foreign_id': 'follow:' +
                                       str(follow_serializer['id']),
                         'follow': follow_serializer,
                         }

        add_activity_to_feed(self.request.user, activity_data)

        '''
            After following, user see activities of followed user
            in his/her timeline.
        '''
        follow_user_timeline(self.request.user,
                             follow_serializer['to_user']['username'])

        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def get_serializer_class(self, *args, **kwargs):
        if self.action == 'create':
            return FollowPostSerializer
        return FollowSerializer

    @action(detail=True, methods=['POST'], name='unfollow',
            serializer_class=None)
    def unfollow(self, request, pk=None):

        following = get_object_or_404(
            Following, pk=pk)
        if following.from_user == self.request.user:
            following.unfollow()
            return Response(status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)


class FollowRequestViewSet(viewsets.ModelViewSet):
    """
    ViewSet for follow request.
    """

    permission_classes = (IsAuthenticatedOrReadOnly,
                          IsRequestSenderOrReceiver)
    queryset = FollowRequest.objects.all()
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['req_to_user']

    def perform_create(self, serializer):
        serializer.save(from_user=self.request.user)

    def create(self, request, *args, **kwargs):
        serializer = FollowRequestPostSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.validated_data['req_from_user'] = self.request.user
        to_user = User.objects.get(
            id=serializer.validated_data['req_to_user']['id'])
        serializer.validated_data['req_to_user'] = to_user
        follow_request = FollowRequest.objects.create(
            **serializer.validated_data)

        ''' Follow Request  Notification '''
        ''' Target --> Follow Request '''
        notify.send(sender=self.request.user,
                    verb="wants to follow you.",
                    recipient=to_user,
                    target=follow_request,
                    description="Follow Request"
                    )

        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def get_serializer_class(self):
        if self.action == 'create':
            return FollowRequestPostSerializer
        return FollowRequestSerializer

    @action(detail=True, methods=['POST'], name='accept_follow_request',
            serializer_class=None)
    def accept_follow(self, request, pk=None):
        follow_request = get_object_or_404(
            FollowRequest, pk=pk)
        if follow_request.req_to_user == self.request.user:
            follow_request.accept()

            ''' Follow Request Accept Notification '''
            ''' Target --> User '''
            user = User.objects.get(id=follow_request.req_from_user)
            notify.send(sender=self.request.user,
                        verb="accepted your follow request.",
                        recipient=user,
                        target=user,
                        description="User"
                        )

            ''' After accepted the follow request, user see
                all activities of followed user in his/her timeline.
            '''
            user_serializer = UserNotificationSerializer(user).data
            follow_user_timeline(self.request.user,
                                 user_serializer['username'])

            return Response(status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)

    @action(detail=True, methods=['POST'], name='reject_follow_request',
            serializer_class=None)
    def reject_follow(self, request, pk):

        follow_request = get_object_or_404(
            FollowRequest, pk=pk)
        if follow_request.req_to_user == self.request.user:
            follow_request.reject()

            return Response(status=status.HTTP_201_CREATED)
        else:
            return Response(data={
                'error': 'Unauthorized'
            }, status=status.HTTP_401_UNAUTHORIZED)



# Generated by Django 3.1.3 on 2020-12-12 14:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0009_auto_20201212_1633'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='followrequest',
            name='unique_request',
        ),
        migrations.AddConstraint(
            model_name='followrequest',
            constraint=models.UniqueConstraint(fields=('req_from_user', 'req_to_user'), name='unique_request'),
        ),
    ]
